#! /usr/bin/env python3

import os
import re
import sys
import copy
import datetime
import argparse
import subprocess
import multiprocessing as mp
from mycotools.lib.kontools import eprint, intro, outro, format_path, multisub, collect_files, findExecs
from mycotools.lib.dbtools import mtdb, masterDB
from mycotools.lib.biotools import dict2fa, fa2dict
from mycotools.acc2fa import famain as acc2fa


def compileBlastCmd( ome, biofile, out_dir, blast_scaf ):
    return blast_scaf + ['-out', out_dir + ome + '.tsv', \
        '-subject', biofile]

def comp_blast_tups( 
    seq_db, blast_type, seq_type, out_dir, 
    biotype, query, hsps = None, max_hits = None,
    evalue = None
    ):

    blast_scaf = [
        blast_type, '-query', query,
        '-outfmt', '6'
        ]
    if evalue:
        blast_scaf.extend(['-evalue', str(evalue)])
    if hsps:
        blast_scaf.extend(['-max_hsps', str(hsps)])
    if max_hits:
        blast_scaf.extend([ '-max_target_seqs', str(max_hits) ])

    blast_cmds = []
    for i, ome in enumerate(seq_db['ome']):
        if seq_db[biotype][i]:
            blast_cmds.append( compileBlastCmd(
                ome, seq_db[biotype][i],
                out_dir, blast_scaf
                ) )
    return blast_cmds

def compileDiamondCmd(ome, dmnd_db, out_dir, blast_scaf):
    return blast_scaf + ['--out', out_dir + ome + '.tsv', \
        '--subject', dmnd_db]

def comp_diamond_tups( 
    seq_db, diamond, blast_type, seq_type, out_dir, 
    biotype, query, hsps = None, max_hits = None,
    evalue = None
    ):

    if not os.path.isdir(out_dir + 'dmnd/'):
        os.mkdir(out_dir + 'dmnd/')
    blast_scaf = [
        diamond, blast_type, '-query', query,
        '--outfmt', '6'
        ]
    if evalue:
        blast_scaf.extend(['--evalue', str(evalue)])
    if hsps:
        blast_scaf.extend(['--max_hsps', str(hsps)])
    if max_hits:
        blast_scaf.extend(['--max_target_seqs', str(max_hits)])

    blast_cmds, db_cmds = [], []
    for i, ome in enumerate(seq_db['ome']):
        if seq_db[biotype][i]:
            db_cmds.append([
                diamond, 'makedb', '--in', seq_db[biotype][i],
                '--db', out_dir + 'dmnd/' + ome, '-p', '2'
                ])
            blast_cmds.append(compileDiamondCmd(
                ome, out_dir + 'dmnd/' + ome,
                out_dir, blast_scaf
                ))
    return db_cmds, blast_cmds



def compMMseqTups( 
    seq_db, out_dir, 
    biotype, query, mmseqs = 'mmseqs',
    evalue = None
    ):

    cmd_scaf = [
        mmseqs, 'easy-search', query, 'tmp'#,
#        '--format-mode', '0'
        ]
    if evalue:
        cmd_scaf.extend(['-e', str(evalue)])
#    if max_hits:
#        cmd_scaf.extend([ '-max_target_seqs', str(max_hits) ])

    search_cmds = []
    for i, ome in enumerate(seq_db['ome']):
        if seq_db[biotype][i]:
            new_cmd = copy.deepcopy(cmd_scaf)
            inputFile = format_path(seq_db[biotype][i])
            outputFile = format_path(out_dir + ome + '.out')
            new_cmd.insert(3, inputFile)
            new_cmd.insert(4, outputFile)
            search_cmds.append(new_cmd)


    return search_cmds



def parseOutput( ome, file_, bitscore = 0, pident = 0 ):

    ome_results = [ome, []]
    if os.path.exists( file_ ):
        with open( file_, 'r' ) as raw:
            raw_data = raw.read()
        data = [x.split('\t') for x in raw_data.split('\n') if x]
        for i in data:
            if int(float(i[-1])) > bitscore and int(1000*float(i[2])) > 100000 * pident:
                ome_results[-1].append(i)

    return ome_results


def compileResults( res_dict, skip = [] ):

    output_res = {}
    for i in res_dict:
        for hit in res_dict[i]:
            query = hit[0]
            subject = hit[1]
            pident = hit[2]
            start = hit[-4]
            end = hit[-3]
            if query not in output_res:
                output_res[query] = {}
            if i not in output_res[query]:
                output_res[query][i] = []

            output_res[query][i].append( [ subject, start, end ] )

    for query in skip:
        del output_res[query]

    return output_res


def compAcc2fa( db, biotype, output_res, subhit = False, skip = None ):

    acc2fa_cmds, db = {}, db.set_index('ome')
    if subhit or biotype == 'assembly':
        for query in output_res:
            if skip:
                continue
            acc2fa_cmds[query] = []
            for i in output_res[query]:
                accs = []
                for hit in output_res[query][i]:
                    accs.append(hit[0] + '[' + hit[1] + '-' + hit[2] + ']')
                if biotype == 'assembly':
                    acc2fa_cmds[query].append(
                        [accs, db[i][biotype]]
                        )
                else:
                    acc2fa_cmds[query].append(
                        [accs, db[i][biotype]]
                        )
    else:
        for query in output_res:
            if skip:
                continue
            acc2fa_cmds[query] = []
            for i in output_res[query]:
                accs = []
                for hit in output_res[query][i]:
                    subject = hit[0]
                    accs.append(subject)
                if accs:
                    try:
                        acc2fa_cmds[query].append( [
                            list(set(accs)), db[i][biotype]
                            ] )
                    except KeyError:
                        eprint('\t' + i + ' not in db')

    return acc2fa_cmds


def prepOutput(out_dir, query):

    out_dir, query = format_path(out_dir), format_path( query )
    if not out_dir.endswith('/'):
        out_dir += '/'
    if not os.path.isdir( out_dir ):
        os.mkdir( out_dir )
    report_dir = out_dir + 'reports/'
    if not os.path.isdir( report_dir ):
        os.mkdir( report_dir )

    return report_dir

def db2searchLog(report_dir, blast, query, max_hits, evalue, bit, pident, out_dir):
    log_list0 = [
        report_dir, blast, query, str(max_hits), str(evalue),
        str(bit), str(pident)
        ]
    
    prev = False    
    log_name = out_dir + '.' + os.path.basename(out_dir[:-1]) + '.log'
    log_list1 = None
    if not os.path.isfile( log_name ):
        with open( log_name, 'w' ) as out:
            out.write( '\n'.join(log_list0) )
    else:
        with open( log_name, 'r' ) as raw:
            log_list1 = raw.read().split('\n')
        if log_list1[:5] == log_list0[:5]:
            eprint('\nPicking up previous run', flush = True)
            prev = True
        else:
            eprint('\nInconsistent run parameters in log, rerunning', flush = True)
            count = 0
            while os.path.isdir( report_dir ):
                count += 1
                report_dir = report_dir[:-1] + str(count) + '/'
            os.mkdir( report_dir )
            with open( log_name, 'w' ) as out:
                out.write( '\n'.join(log_list0) )

    return log_list0, log_list1, prev


def ObyOprep(
    db, report_dir, blast, query, 
    max_hits, evalue, out_dir, bit, pident
    ):

    prev, finished, rundb = False, set(), db

    log_list0, log_list1, prev = db2searchLog(report_dir, blast, query, max_hits, evalue, bit, pident, out_dir)

    reparse = True
    if prev:
        reparse = False
        reports = collect_files(report_dir, 'tsv')
        finished = {
            os.path.basename(x)[:-4] for x in reports \
            if os.path.getsize(x) > 0
            }
        rundb, checkdb = mtdb({}).set_index('ome'), db.set_index('ome')
        for ome, val in checkdb.items():
            if ome not in finished:
                rundb[ome] = val
        rundb = rundb.reset_index()
        if log_list0[-2:] != log_list1[-2:]:
            reparse = True
#        print(rundb, flush = True)
    return rundb


def ObyOsearch(
    db, rundb, blast, seq_type, report_dir, biotype,
    query, hsps, max_hits, evalue, cpus,
    bitscore, pident, diamond
    ):

# NEED to add other threshold options
    if len(rundb) > 0:
        print('\nSearching on an ome-by-ome basis', flush = True)
        if diamond:
            db_tups, search_tups = comp_diamond_tups(
                rundb, diamond, blast, seq_type, report_dir, biotype, query,
                hsps = hsps, max_hits = max_hits, evalue = evalue
                )
            db_outs = multisub(db_tups, processes = cpus)
        elif 'blast' in blast:
            search_tups = comp_blast_tups(
                rundb, blast, seq_type, 
                report_dir, biotype, query, 
                hsps = hsps, max_hits = max_hits, evalue = evalue
                )
        else:
            search_tups = compMMseqTups(
                rundb, report_dir, 
                biotype, query, mmseqs = 'mmseqs',
                evalue = evalue
                )
        search_outs = multisub( search_tups, processes = cpus )
    
    parse_tups = []
    for i, ome in enumerate(db['ome']):
        if db[biotype][i]:
            parse_tups.append( [ome,
                report_dir + ome + '.tsv',
                bitscore, pident ] )

    print('\nParsing reports', flush = True)
    with mp.get_context( 'spawn' ).Pool( processes = cpus ) as pool:
        results = pool.starmap( parseOutput, tuple(parse_tups) )
    results_dict = {x[0]: x[1] for x in results}

    return results_dict


def checkSearchDB(binary = 'blast'):

    db_date = os.path.basename(masterDB())
    if 'blast' in binary:
        search_db = format_path('$MYCOFAA/blastdb/' + db_date + '.00.psd')
        if os.path.isfile(search_db):
            return search_db[:-7]
        elif os.path.isfile(search_db[:-7] + '.psd'):
            return search_db[:-7]
    else:
        search_db = format_path('$MYCOFAA/blastdb/' + db_date.replace('.db','') + '.mmseqs.db')
        if os.path.isfile(search_db):
            return search_db


def dbBlast(
    db_path, blast_type, query, 
    evalue, hsps, cpus,
    report_dir, diamond = False
    ):

    out_file = report_dir + os.path.basename(db_path)[:-3] + '.out'
    if not diamond:
        blast_scaf = [
            blast_type, '-query', query, '-outfmt', '6',
            '-db', db_path, '-num_threads', str(cpus), '-out',
            out_file, '-num_descriptions', str(135904025),
            '-num_alignments', str(135904025)
            ]
        if evalue:
            blast_scaf.extend(['-evalue', str(evalue)])
        if hsps:
            blast_scaf.extend(['-max_hsps', str(hsps)])
    else:
        blast_scaf = [
            diamond, blast_type, '--query', query, '--outfmt', '6',
            '--db', db_path, '-p', str(cpus * 2), '--out',
            out_file, '--num_descriptions', '135904025',
            '--num_alignments', '135904025'
            ]
        if evalue:
            blast_scaf.extend(['--evalue', str(evalue)])
        if hsps:
            blast_scaf.extend(['--max_hsps', str(hsps)])
    
    blast_call = subprocess.call(
        blast_scaf#, stdout = subprocess.PIPE,
    #    stderr = subprocess.PIPE
        )

    return blast_call, out_file


def dbmmseq(
    db_path, query, evalue, cpus, report_dir, mmseqs = "mmseqs", mem = None
    ):

    out_file = report_dir + os.path.basename(db_path)[:-3].replace('.mmseqs','') + '.out'
#    output_str = '"query,target,pident,alen,mismatch,gapopen,qstart,qend,sstart,send,evalue,bits"'
    cmd_scaf = [
        mmseqs, 'easy-search', format_path(query), db_path, out_file, 'tmp',
#        '--format-output', "0", 
        '--threads', str(cpus*2),
        ]
    if mem:
        cmd_scaf.extend(['--split-memory-limit', str(mem)])

    if evalue:
        cmd_scaf.extend(['-e', str(evalue)])

    cmd_call = subprocess.call(
        cmd_scaf#, stdout = subprocess.PIPE,
    #    stderr = subprocess.PIPE
        )

    return cmd_call, out_file


def parseDBout( db, file_, bitscore = 0, pident = 0, max_hits = None ):

    ome_results = {}
    with open( file_, 'r' ) as raw:
        for line in raw:
            data = [x for x in line.rstrip().split('\t')]
            ome = re.search(r'(^[^_]*)', data[1])[1]
            if int(float(data[-1])) > bitscore and int(1000*float(data[2])) > 100000 * pident:
                if ome not in ome_results:
                    ome_results[ome] = []
                ome_results[ome].append(data)

    x_omes = set(db['ome'])
    omes_results = {
        x: ome_results[x] for x in ome_results if x in x_omes
        }

    if max_hits:
        out_results = {}
        for ome in ome_results:
            ome_results[ome].sort(key = lambda x: float(x[-1]), reverse = True)
            out_results[ome] = ome_results[ome][:max_hits]
        return out_results
    else:
        for ome in ome_results:
            ome_results[ome].sort(key = lambda x: float(x[-1]), reverse = True)
            return ome_results


def main( 
    db, blast, query, out_dir, hsps = 1, 
    max_hits = None, evalue = 10, bitscore = 0, 
    pident = 0, mem = None, #coverage = 0,
    cpus = 1, force = False, biotype = None,
    skip = [], diamond = None
    ):

    if blast in {'tblastn', 'blastp'}:
        seq_type = 'prot'
        biotype = 'faa'
        search_db = checkSearchDB()
    elif blast in {'blastx', 'blastn'}:
        seq_type = 'nucl'
        biotype = 'fna'
        search_db = None
    elif blast == 'mmseqs':
        search_db = checkSearchDB('mmseqs')
    else:
        eprint('\nERROR: invalid search binary: ' + blast, flush = True)

    report_dir = prepOutput(out_dir, query)
    if search_db and not force:
        print('\nSearching using MycotoolsDB searchdb', flush = True)
        if 'blast' in blast:
            search_exit, search_output = dbBlast(
                search_db, blast, query, evalue, 
                hsps, cpus, report_dir, diamond = diamond
                )
        else:
            search_exit, search_output = dbmmseq(
                search_db, query, evalue, cpus, report_dir, mmseqs = "mmseqs", mem = mem
                )
        if search_exit:
            eprint('\nERROR: search failed: ' + str(search_exit))
            sys.exit(10)
        results_dict = parseDBout(
            db, search_output, bitscore = bitscore, 
            pident = pident, max_hits = max_hits
            )
    else:
        rundb = ObyOprep(
            db, report_dir, blast, query, 
            max_hits, evalue, out_dir, bitscore, pident
            )
        results_dict = ObyOsearch(
            db, rundb, blast, None, report_dir, biotype,
            query, hsps, max_hits, evalue, cpus,
            bitscore, pident, diamond
            )

    print('\nCompiling fastas', flush = True)
    output_res = compileResults( results_dict, skip )
    acc2fa_cmds = compAcc2fa( db, biotype, output_res, skip = None )
    output_fas = {}
    queryfa = fa2dict(query)
    for query1, cmd in acc2fa_cmds.items():
        output_fas[query1] = {}
        print('\t' + query1, flush = True)
        with mp.get_context('spawn').Pool(processes = cpus) as pool:
            results = pool.starmap( acc2fa, acc2fa_cmds[query1] )
        for x in results:
            output_fas[query1].update(x)
#        output_fas[query1][query1] = queryfa[query1]
        #'\n'.join([dict2fa(x) for x in results])

    return output_fas


if __name__ == '__main__':
# NEED abstract covered portion
# NEED max hits after blast compiled
# NEED hsp option
    mp.set_start_method('spawn')
    parser = argparse.ArgumentParser( 
        description = 'Searches a query against a db and compiles output fastas ' + \
            'for each query. --evalue and --maxhit thresholds are applied at ' + \
            'the search step and are therefore incompatible with --previous. Be ' +
            'aware that db2search.py will preferably search against a pre-indexed database of ' +
            'Mycotools -omes; however, if the database was not generated from ' +
            'the current masterDB (' + os.path.basename(masterDB()) + ') then ' +
            'db2search.py will search on an ome-by-ome basis. In the latter case, ' +
            'e-value is dependent on database size, so it will threshold ' +
            'variably for each ome. If no search database exists for the current masterDB ' +
            'in ' + os.environ['MYCOFAA'] + '/blastdb/ then it is recommended to ' +
            'use the bit score threshold as it is not dependent on database size.'
        )
    parser.add_argument('-s', '--search', required = True, 
        help = 'Search binary {mmseqs, blastn, blastp, tblastn, blastx}')
    parser.add_argument('-d', '--diamond', action = 'store_true',
        help = 'Use diamond version of --search. Not recommended for ' \
             + 'ome-by-ome')
    parser.add_argument('-st', '--sequencetype', help = 'Subject sequence type {aa, nt} for mmseqs')
    parser.add_argument('-q', '--query', required = True, help = 'Query fasta')
    parser.add_argument('-e', '--evalue', type = int,
        help = 'Negative e-value order max, e.g. 2 for 10^-2')
    parser.add_argument('-b', '--bitscore', default = 0,
        type = int, help = 'Bit score minimum')
    parser.add_argument('-i', '--identity', default = 0,
        type = float, help = 'Identity minimum, e.g. 0.6')
    #parser.add_argument( '-c', '--coverage', type = float, help = 'Query coverage +/-, e.g. 0.5' )
    parser.add_argument('-m', '--maxhits', type = int, help = 'Max hits for each organism')
    parser.add_argument('-db', '--mtdb', default = masterDB(), 
        help = 'DEFAULT: masterdb')
    parser.add_argument('-o', '--output', default = os.getcwd() )
    parser.add_argument('--cpu', default = mp.cpu_count(), type = int, help = 'DEFAULT: all')
    parser.add_argument('--ram', help = 'Useful for mmseqs: e.g. 10M or 5G')
    parser.add_argument('-f', '--force', action = 'store_true', help = 'Force ome-by-ome blast')
    args = parser.parse_args()

    db_path = format_path( args.mtdb )
    if args.cpu > mp.cpu_count():
        cpus = mp.cpu_count()
    else:
        cpus = args.cpu
  
    evalue = None 
    if args.evalue:
        evalue = 10**( -args.evalue)

    if args.search == 'mmseqs':
        if not args.sequencetype or args.sequencetype not in {'aa', 'nt'}:
            eprint('\nERROR: -st required for mmseqs', flush = True)
            sys.exit(3)
        elif args.diamond:
            eprint('\nERROR: --diamond cannot be specified with -s mmseqs',
                   flush = True)
            sys.exit(4)
        if args.sequencetype == 'aa':
            biotype = 'faa'
        else:
            biotype = 'fna'
    else:
        biotype = None
        
    start_args = {
        'Search binary': args.search, 'Diamond binary': args.diamond, 'query': format_path(args.query), 
        'database': db_path, 'output': args.output, 'max evalue': evalue,
        'min bitscore': args.bitscore, 'max hits/organism': args.maxhits, #'coverage': args.coverage, 
        'min identity': args.identity, 'cores': cpus, 'force ome-by-ome': args.force
        }

    start_time = intro( 'db2search', start_args )
    date = start_time.strftime('%Y%m%d')
    execs = []
    if args.diamond:
        execs.append('diamond')
    execs.append(args.search)
    findExecs(execs, exit = set(execs))
   
    output_dirPrep = format_path( args.output ) 
    if not output_dirPrep.endswith('/'):
        output_dirPrep = re.sub(r'/+$', '', output_dirPrep)
    output_dir = output_dirPrep + date + '_db2search/'

    db = mtdb(db_path)

    output_fas = main( 
        db, args.search, args.query, output_dir, 
        bitscore = args.bitscore, pident = args.identity, mem = args.ram,
        max_hits = args.maxhits, cpus = cpus, force = args.force, biotype = biotype
        )
    if not os.path.isdir( output_dir + 'fastas/' ):
        os.mkdir( output_dir + 'fastas/' )

    for query in output_fas:
        with open( output_dir + 'fastas/' + query + '.search.fa', 'w' ) as out:
            out.write( dict2fa(output_fas[query]) )

    outro( start_time )



#!/usr/bin/env python3

# NEED to ditch extracthmm and move to simplified output parsing

import os
import re
import sys
import argparse
import datetime
import subprocess
import multiprocessing as mp
from mycotools.extractHmmsearch import main as exHmm, grabNames
from mycotools.extractHmmAcc import main as extr_hmm
from mycotools.db2search import compAcc2fa
from mycotools.acc2fa import famain as acc2fa
from mycotools.lib.kontools import intro, outro, findExecs, eprint, format_path
from mycotools.lib.dbtools import mtdb, masterDB
from mycotools.lib.biotools import dict2fa

def runextractHmmAcc(hmm, accession, output):

    with open(hmm, 'r') as raw:
        hmm_data = raw.read()
    hmm_str = extr_hmm(hmm_data, accession)
    with open(output, 'w') as out:
        out.write(hmm_str)

    return output


def runHmmer(fasta, hmm, output, cpu = 1, binary = 'hmmsearch'):

    hmm_status = subprocess.call([
        binary, '-o', output, '--cpu', str(cpu), hmm, fasta
        ], stdout = subprocess.PIPE,
        stderr = subprocess.PIPE
        )

    return hmm_status


def runExtractHmm(
    hmm_out, top_hits, cov_threshold, evalue, query = True, acc = None
    ):

    with open(hmm_out, 'r') as raw:
        data = raw.read()
    accs = grabNames( data, query = query )
    if len(accs) > 1:
        hmm_data = exHmm(
            data, True, top_hits, cov_threshold, evalue, query = query
            )
    else:
        hmm_data = exHmm(
            data, list(accs)[0], top_hits, cov_threshold, evalue, query =
            query
            )
    
    return hmm_data


def parseHmmData(hmm_data):

    output_res = {}
    for query in hmm_data:
        output_res[query] = {}
        for v in hmm_data[query][1].split('\n'):
            if v.startswith('#') or not v.rstrip():
                continue
            data = [x.rstrip() for x in v.split('\t')]
            seq = data[0]
            start = data[-5]
            end = data[-4]
            ome = re.search(r'(.*?)\_', seq)[1]
            if ome not in output_res[query]:
                output_res[query][ome] = []
            output_res[query][ome].append([seq, start, end])

    return output_res


def runAcc2fa(db, biotype, output_res, subhit = True, cpu = 1):

    acc2fa_cmds = compAcc2fa(db, biotype, output_res, subhit)
    output_fas = {}
    for query in acc2fa_cmds:
        print('\t' + query, flush = True)
        with mp.get_context('spawn').Pool(processes = cpu) as pool:
            results = pool.starmap(acc2fa, acc2fa_cmds[query])
        output_fas[query] = '\n'.join([dict2fa(x) for x in results])

    return output_fas


def outputFas(output_fas, output_dir, fastaname):

    for query in output_fas:
        with open(output_dir + fastaname + '_' + query + '.fa', 'w') as out:
            out.write(output_fas[query])


def main(
    db, binary, fasta_path, hmm_path, out_dir, accession, 
    top_hits = None, cov_threshold = None, evalue = None,
    cpu = 1, accession_search = False, subhit = True 
    ):

    if binary == 'nhmmer':
        biotype = 'fna'
    else:
        biotype = 'faa'

    if accession:
        print('\nExtracting ' + accession, flush = True)
        hmm_path = runextractHmmAcc(
            hmm_path, accession, out_dir + accession + '.hmm'
            )
        if os.path.isfile(accession):
            accession = []
    else:
        print('\nExtracting accessions', flush = True)
        with open(hmm_path, 'r') as raw:
            hmm_data = raw.read()
        accession = []
   
    if cpu > 1:
        hmm_cpu = cpu - 1
    else:
        hmm_cpu = cpu
    hmmer_out = out_dir + 'hmmer.out'
    print('\nRunning ' + binary, flush = True)
    if runHmmer(fasta_path, hmm_path, hmmer_out, cpu = hmm_cpu, binary = binary):
        eprint('\tERROR: ' + binary + ' failed', flush = True)
        sys.exit(2)
    
    print('\nParsing output', flush = True)
    hmm_data = runExtractHmm(
        hmmer_out, top_hits, cov_threshold, evalue, 
        not accession_search, accession
        )
    output_res = parseHmmData(hmm_data)

    print('\nCompiling fastas', flush = True)
    output_fas = runAcc2fa(db, biotype, output_res, subhit = subhit, cpu = cpu)

    return output_fas


if __name__ == '__main__':

    parser = argparse.ArgumentParser( 
        description = 'Runs hmmsearch or nhmmer on a fasta and returns a fasta of hits'
        )
    parser.add_argument('-f', '--fasta', required = True, help = 'Input .fasta')
    parser.add_argument('--hmm', required = True, help = 'Input .hmm')
    parser.add_argument('-b', '--binary', required = True, help = "{'hmmsearch', 'nhmmer'}")
    parser.add_argument('-d', '--mtdb', default = masterDB(), help = 'MycoDB. DEFAULT: master')
    parser.add_argument('-q', '--query', help = 'Query [acc if -a] from .hmm, or new line delimited file')
    parser.add_argument('-c', '--coverage', type = float, help = 'Decimal minimum percent hmm coverage')
    parser.add_argument('-w', '--whole', action = 'store_true', help = 'Extract entire hit sequence' )
    parser.add_argument('-e', '--evalue', type = int, help = 'Maximum evalue threshold 10^(-x)')
    parser.add_argument(
        '-a', '--accession', action = 'store_true', help = 'Extract accession, not query, from .hmm'
        )
    parser.add_argument('-o', '--output', help = 'Output directory')
    parser.add_argument('--cpu', default = 1, type = int)
    args = parser.parse_args()

    if args.binary not in {'hmmsearch', 'nhmmer'}:
        eprint('\nERROR: invalid --binary', flush = True)
        sys.exit(1)
    findExecs([args.binary], exit = {args.binary})

    if args.output:
        out_dir = args.output
    else:
        date = datetime.datetime.today().strftime('%Y%m%d')
        out_dir = os.getcwd() + '/' + date + '_fa2hmm2fa/'
    if not os.path.isdir(out_dir):
        os.mkdir(out_dir)
    out_dir = format_path(out_dir)


    if args.evalue:
        evalue = 10**-(args.evalue)
    else:
        evalue = None

    args_dict = {
        'Fasta': format_path(args.fasta), 'Hmm': format_path(args.hmm), 'Binary': args.binary,
        'MycoDB': format_path(args.mtdb), 'Hmm Acc': args.query, 'Min Coverage': args.coverage,
        'Max E-value': evalue, 'Accessions': args.accession, 'Full Seq': args.whole, 'Output': out_dir, 'CPU': args.cpu
        }
    start_time = intro('fa2hmmer2fa', args_dict)

#    if args.query:
 #       if os.path.isfile(format_path(args.query)):
  #          accs = file2list(args.query)
   #     else:
    #        accs = args.query

    output_fas = main(
        mtdb(format_path(args.mtdb)), args.binary, args.fasta, args.hmm, out_dir, args.query,
        cov_threshold = args.coverage, evalue = args.evalue, cpu = args.cpu,
        accession_search = args.accession, subhit = not args.whole
        )
    fastaname = re.sub(r'\.fa[^\.]*$', '', os.path.basename(os.path.abspath(args.fasta)))
    outputFas(output_fas, out_dir, fastaname)

    outro(start_time)



#! /usr/bin/env python3

import os
import re
import sys
import multiprocessing as mp
from mycotools.lib.biotools import gff2list, list2gff, gff3Comps
from mycotools.lib.kontools import format_path, sys_start

def og2dict(orthogroup_file):

    ome_ogs = {}
    with open(orthogroup_file, 'r') as raw:
        for line in raw:
            data = line.rstrip().split(' ')
            og = int(data[0].replace(':', '').replace('OG',''))
            for gene in data[1:]:
                ome = gene[:gene.index('_')]
                if ome not in ome_ogs:
                    ome_ogs[ome] = {}
                ome_ogs[ome][gene] = og

    return ome_ogs

def sortOGtag(ogtag_dict):
    sort_list = ['K', 'P', 'U', 'C', 'O', 'F', 'G', 'S']

    sorted_dict = {}
    for i in sort_list:
        if i in ogtag_dict:
            sorted_dict[i] = ogtag_dict[i]
            del ogtag_dict[i]
    for i in ogtag_dict:
        sorted_dict[i] = ogtag_dict[i]

    return sorted_dict

def readOGtag(ogtagData):

    ogs = ogtagData.split('|')
    ogtag_dict = {}
    for x in ogs:
        ogtag_info = x.split(':')
        ogtag_dict[ogtag_info[0]] = ogtag_info[1]

    return ogtag_dict

def writeOGtag(ogtag_dict):
    og_str = ''
    for i in ogtag_dict:
        og_str += i + ':' + str(ogtag_dict[i]) + '|'
    return og_str[:-1]

def editOGtag(ogtag_dict, ogtag, og):

    ogtag_dict[ogtag] = og

    new_oginfo = 'OG='
    ogtag_dict = sortOGtag(ogtag_dict)
    for i in ogtag_dict:
        new_oginfo += i + ':' + str(ogtag_dict[i]) + '|'
    new_oginfo = new_oginfo[:-1]

    return new_oginfo

def mycodbOGs(file_path = format_path('$MYCOGFF3/../ogs.tsv'), omes = set()):

    ogInfo_dict = {}
    with open(file_path, 'r') as raw:
        if omes:
            for line in raw:
                og_info = line.rstrip().split('\t')
                gene, ogtag_info = og_info[0], og_info[1]
                if gene[:gene.find('_')] in omes:
                    ogtag_dict = readOGtag(ogtag_info)
                ogInfo_dict[gene] = ogtag_dict
        else:
            for line in raw:
                og_info = line.rstrip().split('\t')
                gene, ogtag_info = og_info[0], og_info[1]
                ogtag_dict = readOGtag(ogtag_info)
                ogInfo_dict[gene] = ogtag_dict

    return ogInfo_dict

def extract_ogs(ogInfo_dict, ogtag):

    og2gene, gene2og = {}, {}
    for gene in ogInfo_dict:
        if ogtag in ogInfo_dict[gene]:
            og = ogInfo_dict[gene][ogtag]
            if og not in og2gene:
                og2gene[og] = []
            og2gene[og].append(gene)
            gene2og[gene] = og

    return og2gene, gene2og

def og2mycoDB(ogInfo_dict, omes = set(), file_path = format_path('$MYCOGFF3/../ogs.tsv')):

    out_list = []
    if os.path.isfile(file_path):
        with open(file_path, 'r') as raw:
            for line in raw:
                data = line.rstrip().split('\t')
                ome = data[0][:data[0].find('_')]
                if ome not in out_list:
                    out_list.append(line.split('\t'))

    for gene in ogInfo_dict:
        out_list.append([gene, writeOGtag(ogInfo_dict[gene])])

    sorted_list = ['\t'.join([str(x) for x in y]) for y in sorted(out_list, key = lambda x: x[0])]
    with open(file_path, 'w') as out:
        out.write('\n'.join(sorted_list))


def og2gff(ogs_dict, gff_path, ogtag):

    gff = gff2list(gff_path)

    for entry in gff:
        if entry['type'] == 'gene':
            gene = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
            if gene in ogs_dict:
                ogSearch = re.search(gff3Comps()['OG'], entry['attributes'])
                if ogSearch:
                    ogtag_dict = readOGtag(ogSearch[1])
                    new_oginfo = editOGtag(ogtag_dict, ogtag, ogs_dict[gene])
                    entry['attributes'] = re.sub(gff3Comps()['OG'], new_oginfo, entry['attributes'])
                else:
                    if not entry['attributes'].endswith(';'):
                        entry['attributes'] += ';'
                    entry['attributes'] += 'OG=' + ogtag + ':' + str(ogs_dict[gene])
                    
    with open(gff_path, 'w') as out:
        out.write(list2gff(gff)) 


def dbMain(og_file, ogtag):
    ome_ogs = og2dict(og_file)
    omes = set(ome_ogs.keys())
    try:
        ogInfo_dict = mycodbOGs(file_path = format_path('$MYCOGFF3/../ogs.tsv'), omes = omes)
    except FileNotFoundError:
        ogInfo_dict = {}
    for ome in ome_ogs:
        for gene in ome_ogs[ome]:
            if gene in ogInfo_dict:
                ogInfo_dict[gene][ogtag] = ome_ogs[ome][gene]
            else:
                ogInfo_dict[gene] = {ogtag: ome_ogs[ome][gene]}

    og2mycoDB(ogInfo_dict, omes = omes, file_path = format_path('$MYCOGFF3/../ogs.tsv'))



def dbMainGff(og_file, ogtag, cpus = 1):

    ome_ogs = og2dict(og_file)

    og2gff_cmds = []
    for ome in ome_ogs:
        og2gff_cmds.append([ome_ogs[ome], format_path('$MYCOGFF3/' + ome + '.gff3'), ogtag])
    with mp.Pool(processes = cpus) as pool:
        pool.starmap(og2gff, og2gff_cmds)


if __name__ == '__main__':

    usage = 'Adds orthogroup tag ("OG=") to mycotoolsDB GFFs from an OrthoFinder Orthogroups.txt file\n' \
        + 'og2mycodb.py <Orthogroups.txt> <TAG>\n\n' \
        + 'Default tags: kingdom [K], phylum [P], subphylum [U], class [C], order [O], family [F], genus [G],' \
        + 'species [S]'
    args = sys_start(sys.argv[1:], usage, 2, files = [sys.argv[1]])
    if len(args) > 2:
        cpus = int(args[2])
    else:
        cpus = 1
    dbMain(format_path(args[0]), args[1])
    sys.exit(0)



#! /usr/bin/env python3

import re
import sys
import argparse
from Bio.Seq import Seq
from mycotools.lib.dbtools import mtdb, masterDB
from mycotools.lib.biotools import fa2dict, gff2list, gff3Comps, dict2fa
from mycotools.lib.kontools import format_path, sys_start, eprint, stdin2str

def sortGene(sorting_group):

    out_group = []
    for entryType in ['gene', 'mrna', 'trna', 'rrna', 'exon', 'cds']:
        todel = []
        for i, entry in enumerate(sorting_group):
            if entry['type'].lower() == entryType:
                out_group.append(entry)
                todel.append(i)
        for i in reversed(todel):
            sorting_group.pop(i)

    for entry in sorting_group:
        out_group.append(entry)

    return out_group 

def sortContig(contigData):

    coordinates = {}
    for gene in contigData:
        t_coords = []
        for entry in contigData[gene]:
            t_coords.extend([int(entry['start']), int(entry['end'])])
        coordinates[gene] = min(t_coords)
    coordinates = {
        k: v for k, v in sorted(coordinates.items(), key = lambda item: item[1])
        }
    outContig = []
    for gene in coordinates:
        outContig.extend(contigData[gene])

    return outContig



def sortGFF(unsorted_gff, idComp):

    sorting_groups, oldGene = {}, None
    for i, entry in enumerate(unsorted_gff):
        seqid = entry['seqid']
        if seqid not in sorting_groups:
            sorting_groups[seqid] = {}
        gene = idComp.search(entry['attributes'])[1]
        gene = re.sub(r'(.*?_\d+).*', r'\1', gene)
        if gene not in sorting_groups[seqid]:
            sorting_groups[seqid][gene] = []
        sorting_groups[seqid][gene].append(entry)

    sortedGff = []
    for seqid in sorting_groups:
        contigData = {}
        for gene in sorting_groups[seqid]:
            contigData[gene] = sortGene(sorting_groups[seqid][gene])
        sortedGff.extend(sortContig(contigData))

    return sortedGff
 
def sortMain(gff):

    id_comp = re.compile(gff3Comps()['id'])
#    crude_sort = sorted( gff, key = lambda x: \
 #       int( re.search(r'ID=([^;]+)', x['attributes'])[1] ))
    gff = sortGFF(gff, id_comp)

    return gff

def grabCDS(gff_dicts, spacer = '\t'):
    """Grab CDSs that are associated with genes. gff_dicts is a
    mycotools.lib.biotools gff2list() list"""

    mrnas, genes, warning, ome = [], [], False, None
    for entry in gff_dicts:
        if entry['type'] == 'mRNA':
            try:
                alias = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
                if not ome:
                    ome = alias[:alias.find('_')]
            except TypeError:
                raise TypeError(str(entry))
            mrnas.extend(alias.split('|')) # account for posttranslational mods
        elif entry['type'] == 'gene':
            try:
                alias = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
            except TypeError:
                if not warning and ome:
                    eprint(spacer + 'WARNING: ' + str(ome) + ' missing aliases', flush = True)
                    warning = True
#                raise TypeError(str(entry))
            genes.extend(alias.split('|')) # account for alternate splicing

    mrna_set = set([x for x in mrnas if x in set(genes)])
    out_cds = []
    for entry in gff_dicts:
        if entry['type'] == 'CDS':
            alias = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
            if alias in mrna_set:
                out_cds.append(entry)


    return out_cds

def order_neg_dict( neg_dict ):

    for seqid in neg_dict:
        for gene in neg_dict[seqid]:
            for cds in neg_dict[seqid][gene]:
                cds.sort( reverse = True )
            neg_dict[seqid][gene] = sorted( 
                neg_dict[seqid][gene], key = lambda x: x[0], reverse = True
                )
    
    return neg_dict

def grabCoords( cdss ):
    
    pos_dict, neg_dict = {}, {}
    for entry in cdss:
        gene = re.search( gff3Comps()['Alias'], entry['attributes'] )[1]
        seqid = entry['seqid']
        if entry['strand'].rstrip() == '+':
            if seqid not in pos_dict:
                pos_dict[seqid], pos_dict[seqid][gene] = {}, []
            elif gene not in pos_dict[seqid]:
                pos_dict[seqid][gene] = []
            pos_dict[seqid][gene].append( [
                int(entry['start']) - 1, int(entry['end'])
                ] )
        elif entry['strand'].rstrip() == '-':
            negseq = seqid
            if seqid not in neg_dict:
                neg_dict[seqid], neg_dict[seqid][gene] = {}, []
            elif gene not in neg_dict[seqid]:
                neg_dict[seqid][gene] = []
            neg_dict[seqid][gene].append( [
                int(entry['start']), int(entry['end'])
                ] )

    return pos_dict, order_neg_dict(neg_dict)

def translatePos( contig_seq, postig_dict ):

    genes_fa_dict = {}
    for gene in postig_dict:
        nt_seq = ''.join([
            contig_seq[x[0]:x[1]] for x in postig_dict[gene]
            ])
        genes_fa_dict[gene] = {
            'sequence': str(Seq(nt_seq).translate()), 'description': ''
            }

    return genes_fa_dict

def translateNeg( rev_seq, negtig_dict ):

    genes_fa_dict = {}
    seqlen = len(rev_seq.rstrip())
    for gene in negtig_dict:
        nt_seq = ''.join([
            rev_seq[seqlen-x[0]:seqlen-x[1]+1] for x in negtig_dict[gene]
            ])
        genes_fa_dict[gene] = {
            'sequence': str(Seq(nt_seq).translate()), 'description': ''
            }

    return genes_fa_dict 

def ntPos( contig_seq, postig_dict ):

    genes_fa_dict = {}
    for gene in postig_dict:
        genes_fa_dict[gene] = {
            'sequence': ''.join([
                contig_seq[x[0]:x[1]] for x in postig_dict[gene]
                ]),
            'description': ''
            }

    return genes_fa_dict


def posPlusMinusCode(genePostig_dict, plusminus, contig_seq, plus = True, minus = True):
    geneDict = {'sequence': '', 'description': ''}
    minimumList = []
    for x in genePostig_dict:
        minimumList.extend(x)
        geneDict['sequence'] += contig_seq[x[0]:x[1]]
    minimum = min(minimumList)
    maximum = max(minimumList)
    if minus:
        if minimum:
            if minimum - plusminus > 0:
                geneDict['sequence'] = contig_seq[minimum-plusminus:minimum] + \
                    geneDict['sequence']
                geneDict['description'] = '-' + str(plusminus)
            else:
                geneDict['sequence'] = contig_seq[0:minimum] + \
                    geneDict['sequence']
                geneDict['description'] = '-' + str(minimum)
    if plus:
        geneDict['sequence'] += contig_seq[maximum+1:maximum+plusminus+1]
        if len(contig_seq) < maximum + plusminus:
            geneDict['description'] += '+' + str(len(contig_seq) - maximum)
        else:
            geneDict['description'] += '+' + str(plusminus)
    return geneDict


def posPlusMinus(genePostig_dict, plusminus, contig_seq, plus = True, minus = True):
    minimumList = []
    for x in genePostig_dict:
        minimumList.extend(x)
    minimum = min(minimumList)
    maximum = max(minimumList)

    geneDict = {'sequence': '', 'description': ''}
    if minus:
        if minimum - plusminus > 0:
            geneDict['sequence'] = contig_seq[minimum-plusminus:maximum+1]
            if plusminus:
                geneDict['description'] = '-' + str(plusminus)
        else:
            geneDict['sequence'] = contig_seq[:maximum+1]
            if plusminus:
                geneDict['description'] = '-' + str(minimum)
    if plus:
        geneDict['sequence'] += contig_seq[maximum+1:maximum+1+plusminus]
        if plusminus:
            if maximum + plusminus > len(contig_seq):
                if geneDict['description']:
                    geneDict['description'] += ';'
                geneDict['description'] += '+' + str(len(contig_seq) - maximum)
            else:
                if geneDict['description']:
                    geneDict['description'] += ';'
                geneDict['description'] += '+' + str(plusminus)
    return geneDict

def ntPosNoncode(contig_seq, postig_dict, plusminus = 0):

    genes_fa_dict = {}
    for i, gene in enumerate(list(postig_dict.keys())):
        genes_fa_dict[gene] = {'sequence': '', 'description': ''}
        if plusminus:
            genes_fa_dict[gene] = posPlusMinus(postig_dict[gene], plusminus, contig_seq)
        else:
            minimumList = []
            for x in postig_dict[gene]:
                minimumList.extend(x)
            minimum = min(minimumList)
            maximum = max(minimumList)
    
            genes_fa_dict[gene]['sequence'] = contig_seq[minimum:maximum]
   
    return genes_fa_dict

def negPlusMinusCode(geneNegtig_dict, seqlen, plusminus, rev_seq, plus = True, minus = True):
    minimumList, geneDict = [], {'sequence': '', 'description': ''}
    for x in geneNegtig_dict:
        minimumList.extend(x)
        geneDict['sequence'] += rev_seq[seqlen-x[0]:seqlen-x[1]+1]
    minimum = min(minimumList)
    maximum = max(minimumList)
    if minus:
        if maximum + plusminus < seqlen:
            geneDict['sequence'] = \
                rev_seq[seqlen-maximum-plusminus:seqlen-maximum] + geneDict['sequence']
            geneDict['description'] = '-' + str(plusminus) + ';+'
        else:
            geneDict['sequence'] = \
                rev_seq[:seqlen-maximum] + geneDict['sequence']
            geneDict['description'] = '-' + str(seqlen-maximum) + ';+'
    if plus:
        if minimum - plusminus > 0:
            geneDict['sequence'] += rev_seq[seqlen-minimum+1:seqlen-minimum+plusminus+1]
            geneDict['description'] += str(plusminus)
        else:
            geneDict['sequence'] += rev_seq[seqlen-minimum+1:]
            geneDict['description'] += str(seqlen-minimum)   
    return geneDict

def negPlusMinus(neg_gene, seqlen, plusminus, rev_seq, plus = True, minus = True):
    minimumList = []
    for x in neg_gene:
        minimumList.extend(x)
    minimum = min(minimumList)
    maximum = max(minimumList)

    geneDict = {'sequence': '', 'description': ''}
    if minus:
        if maximum + plusminus < seqlen:
            geneDict['sequence'] = \
                rev_seq[seqlen-maximum-plusminus:seqlen-maximum]
            if plusminus:
                geneDict['description'] = '-' + str(plusminus)
        else:
            geneDict['sequence'] = \
                rev_seq[:seqlen-maximum]
            if plusminus:
                geneDict['description'] = '-' + str(seqlen-maximum)
    if plus:
        if minimum - plusminus > 0:
            geneDict['sequence'] += rev_seq[seqlen-maximum:seqlen-minimum+plusminus+1]
            if plusminus:
                if geneDict['description']:
                    geneDict['description'] += ';'
                geneDict['description'] += '+' + str(plusminus)
        else:
            geneDict['sequence'] += rev_seq[seqlen-maximum+1:]
            if plusminus:
                if geneDict['description']:
                    geneDict['description'] += ';'
                geneDict['description'] += '+' + str(seqlen-minimum)   
    return geneDict


def ntNeg( rev_seq, negtig_dict, plusminus = 0 ):

    genes_fa_dict = {}
    seqlen = len(rev_seq.rstrip())
    for gene in negtig_dict:
        genes_fa_dict[gene] = {
            'sequence': ''.join([
                rev_seq[seqlen-x[0]:seqlen-x[1]+1] for x in negtig_dict[gene]
                ]),
            'description': ''
            }

    return genes_fa_dict 

def ntNegNoncode(rev_seq, negtig_dict, plusminus = 0):

    genes_fa_dict = {}
    seqlen = len(rev_seq.rstrip())
    for gene in negtig_dict:
        if plusminus:
            genes_fa_dict[gene] = negPlusMinus(negtig_dict[gene], plusminus, seqlen, rev_seq)
        else:
            minimumList = []
            for x in negtig_dict[gene]:
                minimumList.extend(x)
            minimum = min(minimumList)
            maximum = max(minimumList)

            genes_fa_dict[gene] = {
                'sequence': rev_seq[seqlen - maximum:seqlen - minimum + 1],
                'description': ''
                }

    return genes_fa_dict


def ntmain(gff_dicts, assem_dict, coding = True, 
           flanks = True, fullRegion = False, plusminus = 0,
           spacer = '\t'):
    
    if fullRegion:
        flanks, coding = True, False
    contig_dict, contig_info, genes_fa_dict, geneOrder = {}, {}, {}, {}
    cdss = grabCDS( sortMain(gff_dicts), spacer = spacer )
    for cds in cdss:
        seqid = cds['seqid']
        if seqid not in contig_dict:
            contig_dict[seqid] = []
            geneOrder[seqid] = []
        contig_dict[seqid].append(cds)
        gene = re.search(gff3Comps()['Alias'], cds['attributes'])[1]
        if gene not in set(geneOrder[seqid]):
            geneOrder[seqid].append(gene)
    for seqid in contig_dict:
        contig_info[seqid] = [(
            re.search(gff3Comps()['Alias'], contig_dict[seqid][0]['attributes'])[1],
            contig_dict[seqid][0]['strand']
            ),(
            re.search(gff3Comps()['Alias'], contig_dict[seqid][-1]['attributes'])[1],
            contig_dict[seqid][-1]['strand']
            )]

    pos_dict, neg_dict = grabCoords( cdss )

    contig_fa_dict, startFlanks, endFlanks = {seqid: {} for seqid in contig_info}, {}, {}
    if flanks:
        if coding:
            for seqid in contig_info:
                startGene, startStrand = contig_info[seqid][0][0], contig_info[seqid][0][1]
                endGene, endStrand = contig_info[seqid][1][0], contig_info[seqid][1][1]
                if startGene == endGene:
                    if startStrand == '+':
                        startFlanks[seqid] = posPlusMinusCode(
                            pos_dict[seqid][startGene],
                            plusminus,
                            assem_dict[seqid]['sequence'],
                            )
                        del pos_dict[seqid][startGene]
                    else:
                        rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                        startFlanks[seqid] = negPlusMinusCode(
                            neg_dict[seqid][startGene], len(rev_comp), plusminus, rev_comp
                            )
                        del neg_dict[seqid][startGene]
                    continue
                if startStrand == '+':
                    startFlanks[seqid] = posPlusMinusCode(
                        pos_dict[seqid][startGene],
                        plusminus,
                        assem_dict[seqid]['sequence'],
                        plus = False
                        )
                    del pos_dict[seqid][startGene]
                else:
                    rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                    startFlanks[seqid] = negPlusMinusCode(
                        neg_dict[seqid][startGene], len(rev_comp), plusminus, rev_comp, plus = False
                        )
                    del neg_dict[seqid][startGene]
                if endStrand == '+':
                    endFlanks[seqid] = posPlusMinusCode(
                        pos_dict[seqid][endGene],
                        plusminus,
                        assem_dict[seqid]['sequence'],
                        minus = False
                        )
                    del pos_dict[seqid][endGene]
                else:
                    rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                    endFlanks[seqid] = negPlusMinusCode(
                        neg_dict[seqid][endGene], len(rev_comp), plusminus, rev_comp, minus = False
                        )
                    del neg_dict[seqid][endGene]
        else:
            for seqid in contig_info:
                startGene, startStrand = contig_info[seqid][0][0], contig_info[seqid][0][1],
                endGene, endStrand = contig_info[seqid][1][0], contig_info[seqid][1][1]
                      
                if startGene == endGene:
                    if startStrand == '+':
                        startFlanks[seqid] = posPlusMinus(
                            pos_dict[seqid][startGene],
                            plusminus,
                            assem_dict[seqid]['sequence'],
                            )
                        del pos_dict[seqid][startGene]
                    else:
                        rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                        startFlanks[seqid] = negPlusMinus(
                            neg_dict[seqid][startGene], len(rev_comp), plusminus, rev_comp
                            )
                        del neg_dict[seqid][startGene]
                    continue

                elif fullRegion:
                    region = []
                    if startStrand == '+':
                        region.append(min(pos_dict[seqid][startGene]))
                    else:
                        region.append(min(neg_dict[seqid][startGene]))
                    if endStrand == '+':
                        region.append(max(pos_dict[seqid][endGene]))
                    else:
                        region.append(max(neg_dict[seqid][endGene]))
                    name = startGene + '-' + endGene
                    genes_fa_dict[name + '_sense'] = posPlusMinus(
                        region, plusminus, assem_dict[seqid]['sequence']
                        )
                    rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                    genes_fa_dict[name + '_antisense'] = negPlusMinus(
                        region, len(rev_comp), plusminus, rev_comp
                        )

                
                if startStrand == '+':
                    startFlanks[seqid] = posPlusMinus(
                        pos_dict[seqid][startGene],
                        plusminus,
                        assem_dict[seqid]['sequence'],
                        plus = False
                        )
                    del pos_dict[seqid][startGene]
                else:
                    rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
                    startFlanks[seqid] = negPlusMinus(
                        neg_dict[seqid][startGene], len(rev_comp), plusminus, rev_comp, plus = False
                        )
                    del neg_dict[seqid][startGene]
                if endStrand == '+':
                    endFlanks[seqid] = posPlusMinus(
                        pos_dict[seqid][endGene],
                        plusminus,
                        assem_dict[seqid]['sequence'],
                        minus = False
                        )
                    del pos_dict[seqid][endGene]
                else:
                    rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                    endFlanks[seqid] = negPlusMinus(
                        neg_dict[seqid][endGene], len(rev_comp), plusminus, rev_comp, minus = False
                        )
                    del neg_dict[seqid][endGene]

        if coding:
            for contig in contig_fa_dict:
                if contig in pos_dict:
                    contig_fa_dict[contig] = { 
                        **ntPos(assem_dict[contig]['sequence'], pos_dict[contig]), 
                        **contig_fa_dict[contig]
                        }
                if contig in neg_dict:
                    rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
                    contig_fa_dict[contig] = {
                        **ntNeg(rev_comp, neg_dict[contig]),
                        **contig_fa_dict[contig]
                        }
        elif not fullRegion:
            for contig in contig_fa_dict:
                if contig in pos_dict:
                    contig_fa_dict[contig] = { 
                        **ntPosNoncode(assem_dict[contig]['sequence'], pos_dict[contig]), 
                        **contig_fa_dict[contig]
                        }
                if contig in neg_dict:
                    rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
                    contig_fa_dict[contig] = {
                        **ntNegNoncode(rev_comp, neg_dict[contig]),
                        **contig_fa_dict[contig]
                        }
    if not fullRegion:
        for contig in geneOrder:
            if contig in startFlanks:
                genes_fa_dict[geneOrder[contig][0]] = startFlanks[contig]
            for gene in geneOrder[contig][1:-1]:
                genes_fa_dict[gene] = contig_fa_dict[contig][gene]
            if contig in endFlanks:
                genes_fa_dict[geneOrder[contig][-1]] = endFlanks[contig]
    
    return genes_fa_dict

def aamain( gff_dicts, assem_dict, spacer = '\t' ):
    
    cdss = grabCDS( gff_dicts, spacer )
    pos_dict, neg_dict = grabCoords( cdss )

    genes_fa_dict = {}
    for contig in pos_dict:
        genes_fa_dict = { 
            **translatePos(assem_dict[contig]['sequence'], pos_dict[contig]), 
            **genes_fa_dict 
            }
    for contig in neg_dict:
        rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
        genes_fa_dict = {
            **translateNeg(rev_comp, neg_dict[contig]),
            **genes_fa_dict
            }

    return genes_fa_dict

            
if __name__ == '__main__':

    parser = argparse.ArgumentParser(
       description = 'Inputs MycoDB compatible gff3, assembly (optional), ' + \
           'and outputs nucleotides/proteins. Use an abstracted gene gff ' + \
           '(acc2gff.py) to only output a smaller set of gene(s).'
       )
    parser.add_argument('-g', '--gff', help = '"-" for stdin', required = True)
    parser.add_argument('-n', '--nucleotide', action = 'store_true')
    parser.add_argument('-p', '--protein', action = 'store_true')
    parser.add_argument('-a', '--assembly')
    parser.add_argument('-i', '--intergenic', action = 'store_true', help = '-n only')
    parser.add_argument('-nc', '--noncoding', action = 'store_true', help = '-n only')
    parser.add_argument('-pm', '--plusminus', default = 0, type = int, help = '-n only')
    parser.add_argument('-af', '--all_flanks', action = 'store_true', help = '-n and -nc only')
    args = parser.parse_args()

    if args.gff == '-':
        data = stdin2str()
        input_gff = gff2list(data, path = False)
    else:
        input_gff = gff2list(format_path(args.gff))
    if args.assembly:
        assembly_dicts = {'input': fa2dict( format_path(args.assembly) )}
        gff_dicts = {'input': input_gff}
    else:
        db = mtdb(masterDB()).set_index('ome')
        gff_dicts, assembly_dicts = {}, {}
        try:
            for line in input_gff:
                gene = re.search(gff3Comps()['Alias'], line['attributes'])[1]
                ome = re.search( r'(.*?)_', gene )[1]
                if ome not in gff_dicts:
                    gff_dicts[ome] = []
                    assembly_dicts[ome] = fa2dict(db[ome]['fna'])
                gff_dicts[ome].append(line)
        except IndexError:
            eprint('\nERROR: ' + args.gff + ' is incompatible with MycotoolsDB', flush = True)
            sys.exit(1)
        
    if args.protein:
        for ome in gff_dicts:
            print(dict2fa(aamain(gff_dicts[ome], assembly_dicts[ome])), flush = True)
    else:
        for ome in assembly_dicts:
            print(dict2fa(
                ntmain(gff_dicts[ome], assembly_dicts[ome], not args.noncoding, not args.all_flanks, args.intergenic, args.plusminus)
                ), flush = True)

    sys.exit(0)



#! /usr/bin/env python3

import os
import sys
import time
import getpass
from Bio import Entrez
from mycotools.lib.kontools import file2list, eprint


if __name__ == '__main__':

    usage = "\nInput NCBI accession or new line delimitted file of accessions and optionally the column name.\n"
    usage += 'ncbiAccs2fa.py <ACC> <COLNAME>\n'
    if "-h" in sys.argv or "--help" in sys.argv:
        eprint(usage, flush = True)
        sys.exit( 1 )
    elif len( sys.argv ) < 2:
        eprint(usage, flush = True)
        sys.exit( 1 )
    elif len( sys.argv ) <= 3:
        if os.path.isfile( sys.argv[1] ):
            if len( sys.argv ) == 3:
                accs = file2list( sys.argv[1], sep = '\t', col = sys.argv[2] )
            else:
                accs = file2list( sys.argv[1] )
        else:
            accs = [ sys.argv[1] ]
    else:
        eprint(usage, flush = True)
        sys.exit( 1 )

    email = input( "\nInput NCBI login email: " )
    limit = 3
    Entrez.email = email
    if len( accs ) > 3:
        api = getpass.getpass( prompt = "NCBI API key (leave blank if none): " )
        if api != "":
            Entrez.api_key = api
            limit = 10

    eprint(flush = True)

    count, amount, out_str = 0, 1, ''
    for acc in accs:
        count += 1
        if count >= limit:
            count = 0
            time.sleep( 1 )
        eprint( acc , flush = True)
        while True:
            try:
                handle = Entrez.efetch( db = "protein", id=acc, retmode = "xml" )
                records = Entrez.read(handle)
                out_str += ">" + acc + ' ' + records[0]['GBSeq_organism'].replace(' ','_') + '\n' + records[0]["GBSeq_sequence"].upper() + '\n'
                break
            except:
                time.sleep( 1)

    eprint(flush = True)
    with open( sys.argv[1] + '.retr.fa', 'w' ) as out:
        out.write( out_str )

    sys.exit( 0 )



#! /usr/bin/env python3

import os
import re
import sys
import argparse
import multiprocessing as mp
from mycotools.lib.biotools import gff2list, list2gff
from mycotools.lib.dbtools import mtdb, masterDB
from mycotools.lib.kontools import format_path, stdin2str

def grabGffAcc( gff_list, acc, term = 'Alias=' ):
    '''grab acc from alias'''

    alias = term + acc
    alias_on = alias + ';'
#    elif ' alias "' in gff_list[0]['attributes']:
 #       alias = 'alias "' + acc + '"'
  #      alias_on = 'alias "' + acc + '"'
    out_list = [ 
        x for x in gff_list \
        if x['attributes'].endswith(alias) or \
            alias_on in x['attributes'] 
        ]
    return out_list


def grabGffAccs(gff_list, acc_list, ome = None):

    aliases, ends = set(), set()
    for acc in acc_list:
        aliases.add('Alias=' + acc)
        ends.add('Alias=' + acc + ';')
    out_list = []
    for i in gff_list:
        if any(i['attributes'].endswith(x) for x in aliases):
            out_list.append(i)
        elif any(x in i['attributes'] for x in ends):
            out_list.append(i)

    if ome:
        return out_list, ome
    else:
        return out_list


def gffMain(gffData, accs):
    accGffs = {}
    if isinstance(gffData, str):
        gff = gff2list(gffData)
    else:
        gff = gffData
    for acc in accs:
        accGffs[acc] = grabGffAcc(gff, acc)
    return accGffs

def dbMain(db, accs, cpus = 1):

    omes = set([x[:x.find('_')] for x in accs])
    db = db.set_index('ome')
    grabAcc_cmds = []
    for ome in list(omes):
        omeAccs = [acc for acc in accs if acc.startswith(ome + '_')]
        gff_list = gff2list(db[ome]['gff3'])
#        grabAcc_res = [grabGffAccs(gff_list, omeAccs, ome)]
        grabAcc_cmds.append([gff_list, omeAccs, ome])

    with mp.Pool(processes = cpus) as pool:
        grabAcc_res = pool.starmap(grabGffAccs, grabAcc_cmds)

    gffs = {}
    for res in grabAcc_res:
        gffs[res[1]] = res[0]

    return gffs


if __name__ == '__main__':

#    mp.set_start_method('spawn')
    parser = argparse.ArgumentParser( 
        description = 'Inputs gff or database and new line delimitted file of accessions.'
        )
    parser.add_argument('-a', '--accession', help = '"-" for stdin')
    parser.add_argument('-i', '--input', help = 'File with accessions')
    parser.add_argument('-g', '--gff', help = 'GFF3 input')
    parser.add_argument('-c', '--column', default = 1, 
        help = 'Accssions column for -i (1 indexed). DEFAULT: 1')
    parser.add_argument('-o', '--ome', action = 'store_true', help = 'Output files by ome code')
    parser.add_argument('-d', '--mtdb', default = masterDB(), help = 'mycodb DEFAULT: master')
    parser.add_argument('--cpu', type = int, default = mp.cpu_count())
    args = parser.parse_args()

    if args.input:
        input_file = format_path( args.input )
        with open(input_file, 'r') as raw:
            accs = [x.rstrip().split('\t')[args.column-1] for x in raw]
    elif not args.accession:
        raise ValueError('need input file or accession')
    else:
        if '-' == args.accession:
            data = stdin2str()
            accs = data.split()
        else:
            accs = [args.accession]

    if args.cpu < mp.cpu_count():
        cpu = args.cpu
    else:
        cpu = mp.cpu_count()

    db_path = format_path( args.mtdb )
    if not args.gff:
        db = mtdb( format_path(args.mtdb) )
        gff_lists = dbMain( db, accs, cpus = args.cpu )
    else:
        gff_path = format_path( args.gff )
        gff_lists = gffMain(gffData, accs)

    if args.accession:
        print( list2gff(gff_lists[list(gff_lists.keys())[0]]).rstrip() , flush = True)
    elif args.ome:
        output = mkOutput(os.getcwd() + '/', 'acc2gff')
        for ome in gff_strs:
            if gff_lists[ome]:
                with open( output + ome + '.accs.gff3', 'w' ) as out:
                    out.write( list2gff(gff_lists[ome]) )
            else:
                eprint('ERROR: ' + ome + ' failed, no accessions retrieved', flush = True)
    else:
        out_str = ''
        for ome in gff_lists:
            if gff_lists[ome]:
                out_str += list2gff(gff_lists[ome]) + '\n'
            else:
                eprint('ERROR: ' + ome + ' does not have accession', flush = True)
        print(out_str)

    sys.exit( 0 )



#! /usr/bin/env python3

import os
import sys
import copy
import random
import argparse
from collections import defaultdict
from mycotools.lib.kontools import file2list, intro, outro, format_path, eprint
from mycotools.lib.dbtools import mtdb, masterDB

def extract_unique(db, allowed = 1, sp = True):
    keys = copy.deepcopy(list(db.keys()))
    random.shuffle(keys)
    prep_db0 = {x: db[x] for x in keys}
    prep_db1 = mtdb().set_index('ome')
    if sp:
        found = defaultdict(int)
        for ome, row in prep_db0.items():
            name = row['taxonomy']['species']
            found[name] += 1
            if found[name] <= allowed:
                prep_db1[ome] = row
    else:
        found = set()
        for ome, row in prep_db0.items():
            name = row['taxonomy']['species'] + ' ' + row['strain']
            if name not in found:
                prep_db1[ome] = row
            found_prep = list(found)
            found_prep.append(name)
            found = set(found_prep)
    return prep_db1

def extract_tax(db, lineages, classification):
    if isinstance(lineages, str):
        lineages = [lineages]
    lineages = set(x.lower() for x in lineages)
    new_db = mtdb().set_index()
    for ome in db:
        try:
            if db[ome]['taxonomy'][classification].lower() in lineages:
                new_db[ome] = db[ome]
        except KeyError:
            pass

    return new_db

def extract_ome(db, omes):
    new_db = mtdb().set_index()
    for i in db:
        if i in omes:
            new_db[i] = db[i]
    return new_db

def extract_source(db, source):
    return mtdb({ome: row for ome, row in db.items() \
                   if row['source'].lower() == source.lower()}, 
                   index = 'ome')

def extract_pub(db):
    new_db = mtdb().set_index()
    for ome, row in db.items():
        if row['published']:
            new_db[ome] = row
    return new_db

def main( 
    db, unique_strains = False, unique_species = 0, 
    lineage_list = [], rank = None, omes_set = set(),
    source = None, nonpublished = False, inverse = False
    ):

    db = db.set_index('ome')
    if unique_strains:
        db = extract_unique(db, sp = False)
    if unique_species > 0:
        db = extract_unique(db, unique_species, sp = True)

    # extract each taxonomic entry based on the classification specified
    if lineage_list:
        new_db = extract_tax(db, lineage_list, rank)
    # if an ome list is specified then open it, store each entry in a list and pull each ome
    elif omes_set:
        new_db = extract_ome(db, omes)
    # if none of these are specified then create a `new_db` variable to work for later
    else:
        new_db = db

    # if there is a source specified, extract it or the opposite
    if source:
        new_db = extract_source(new_db, source)

    # if you want publisheds, then just pull those out 
    if not nonpublished:
        new_db = extract_pub(new_db)

    if inverse:
        new_omes = set(new_db.keys())
        inv_db = mtdb().set_index()
        for ome, row in db.items():
            if ome not in new_omes:
                inv_db[ome] = row
        return inv_db.reset_index()
    else:
        return new_db.reset_index()


if __name__ == '__main__':

    parser = argparse.ArgumentParser( description = \
       'Extracts a MycotoolsDB from arguments. E.g.\t`extractDB.py ' + \
       '-l Atheliaceae -r family`' )
    parser.add_argument('-l', '--lineage')
    parser.add_argument('-r', '--rank', help = "Taxonomy rank")
    parser.add_argument('-s', '--source')
    parser.add_argument('-n', '--nonpublished', action = 'store_true', 
        help = 'Include restricted')
    parser.add_argument('-u', '--unique_strain', action = 'store_true')
    parser.add_argument('-a', '--allowed_sp', default = 0, type = int, 
        help = 'Replicate species allowed' )
    parser.add_argument('-i', '--inverse', action = 'store_true', 
        help = 'Inverse [source|lineage(s)|nonpublished]')
    parser.add_argument('-ol', '--ome', help = "File w/list of omes" )
    parser.add_argument('-ll', '--lineages', help = 'File w/list of lineages (same rank)' )
    parser.add_argument('--headers', action = 'store_true')
    parser.add_argument('-', '--stdin', action = 'store_true')
    parser.add_argument('-d', '--mtdb', default = masterDB())
    parser.add_argument('-o', '--output' )
    args = parser.parse_args()
    db_path = format_path( args.mtdb )


    if (args.lineage or args.lineages) and not args.rank:
        eprint('\nERROR: need rank for lineage(s)', flush = True)
        sys.exit(5)
    elif args.lineage or args.lineages:
        eprint("\nWARNING: extracting taxonomy is subject to taxonomic error and errors in NCBI's hiearchy\n")

    output = 'stdout'
    if args.output:
        output = format_path( args.output )
        if not output.endswith( '/' ):
            tag = ''
                       
            if args.lineage:
                tag += '_' + args.lineage
            if args.lineages:
                tag += '_taxonomy'
            if args.source:
                tag += args.source.lower()
            if not args.nonpublished:
                tag += '_pub'
            output += '/' + os.path.basename( db_path ) + tag
        
    args_dict = {
        'database': db_path,
        'output': output,
        'lineage': args.lineage,
        'lineages list': args.lineages,
        'rank': args.rank,
        'ome_list': args.ome,
        'Source': args.source,
        'unique': args.allowed_sp,
        'nonpublished': args.nonpublished,
        'inverse': args.inverse,
        'headers': bool(args.headers)
    	}
#    start_time = intro( 'Abstract Database', args_dict, stdout = args.output )

    if args.stdin:
        data = ''
        for line in sys.stdin:
            data += line.rstrip() + '\n'
        data = data.rstrip()
        db = mtdb(data, stdin=True)
    else:
       db = mtdb( db_path )

    if args.ome:
        omes = set(file2list(format_path(args.ome)))
    else:
        omes = set()

    if args.lineages:
        lineage_list = file2list(format_path(lineage_list))
    elif args.lineage:
        lineage_list = [args.lineage]
    else:
        lineage_list = []

    new_db = main( 
        db, rank = args.rank, lineage_list = lineage_list,
        omes_set = omes, source = args.source, unique_species = args.allowed_sp, 
        nonpublished = args.nonpublished, inverse = args.inverse
        )
    if args.output:
        new_db.df2db( output )
    else:
        new_db.df2db( headers = bool(args.headers) )

 #   outro( start_time, stdout = args.output )
    sys.exit(0)



#! /usr/bin/env python3

# NEED a db check to ensure the log is relevant to the input

import os
import re
import sys
import gzip
import time
import urllib
import requests
import argparse
import subprocess
import numpy as np
import pandas as pd
from Bio import Entrez
from datetime import datetime
from mycotools.lib.kontools import intro, outro, format_path, prep_output, eprint, vprint, findExecs
from mycotools.lib.dbtools import log_editor, loginCheck, mtdb, read_tax

def db2df(data, stdin = False):
    import pandas as pd, pandas
    columns = mtdb.columns
    if isinstance(data, mtdb):
        db_df = pd.DataFrame(data.reset_index())
    elif not stdin:
        data = format_path( data )
        db_df = pd.read_csv( data, sep='\t' )
        if 'ome' not in set( db_df.columns ) and 'assembly_acc' not in set( db_df.columns ):
            db_df = pd.read_csv( data, sep = '\t', header = None )
    else:
        db_df = pd.read_csv( StringIO(data), sep='\t' )
        if 'ome' not in set( db_df.columns ) and 'assembly_acc' not in set( db_df.columns ):
            db_df = pd.read_csv( StringIO(data), sep = '\t', header = None )

    db_df = db_df.fillna('')

    if len(db_df.keys()) == 16: # legacy conversion TO BE DEPRECATED
        eprint('\tWARNING: Legacy MycotoolsDB format will be removed in the future.', flush = True)
        db_df.columns = [
            'ome', 'genus', 'species', 'strain', 'version',
            'biosample', 'fna', 'faa', 'gff3', 'taxonomy', 'ecology',
            'eco_conf',
            'source', 'published', 'assembly_acc', 'acquisition_date'
            ]
    else:
        db_df.columns = columns
    for i, row in db_df.iterrows():
        db_df.at[i, 'taxonomy'] = read_tax(
            row['taxonomy']
            )
        db_df.at[i, 'taxonomy']['genus'] = row['genus']
        db_df.at[i, 'taxonomy']['species'] = \
            row['genus'] + ' ' + row['species']
        # if malformatted due to decreased entries in some lines, this will raise an IndexError
        if row['fna'] or row['fna'] == row['ome'] + '.fna': # abbreviated line w/o file coordinates
           db_df.at[i, 'fna'] = os.environ['MYCOFNA'] + row['ome'] + '.fna'
           db_df.at[i, 'faa'] = os.environ['MYCOFAA'] + row['ome'] + '.faa'
           db_df.at[i, 'gff3'] = os.environ['MYCOGFF3'] + row['ome'] + '.gff3'
        else: # has file coordinates
           db_df.at[i, 'fna'] = format_path(row['fna'])
           db_df.at[i, 'faa'] = format_path(row['faa'])
           db_df.at[i, 'gff3'] = format_path(row['gff3'])

    if len(db_df) == 16: # LEGACY conversion to be deprecated
        del db_df['ecology']
        del db_df['eco_conf']

    return db_df


def prepare_folders( output_path, gff, prot, assem, transcript ):

    file_types = []
    if assem:
        if not os.path.exists(output_path + 'assembly'):
            os.mkdir(output_path + 'assembly')
        file_types.append( 'assembly' )
    if gff:
        if not os.path.exists(output_path + 'gff3'):
            os.mkdir(output_path + 'gff3')
        file_types.append( 'gff3' )
    if prot:
        if not os.path.exists(output_path + 'proteome'):
            os.mkdir(output_path + 'proteome')
        file_types.append( 'proteome' ) 
    if transcript:
        if not os.path.exists(output_path + 'transcript'):
            os.mkdir(output_path + 'transcript')
        file_types.append( 'transcript' )

    return file_types  


def compile_log( output_path, remove = False ):

    ass_prots = {}
    if not os.path.isfile( output_path + 'ncbiDwnld.log' ):
        with open( output_path + 'ncbiDwnld.log', 'w' ) as out:
            out.write('#ome\tassembly_acc\tassembly\tproteome\tgff3\ttranscript\t' + \
            'assemMD5\tprotMD5\tgff3MD5\ttransMD5\tgenome_id(s)')

        # too risky, too many things can go wrong and then users would be in a
        # loop, but necessary for huge downloads
    else:
        with open( output_path + 'ncbiDwnld.log', 'r' ) as raw:
            for line in raw:
                if not line.startswith('#'):
                    data = [x.rstrip() for x in line.split('\t')]
                    while len(data) < 10:
                        data.append('')
                    ass_prots[data[0]] = { 
                        'assembly_acc': str(data[1]), 'assembly': str(data[2]), 
                        'proteome': str(data[3]), 'gff3': str(data[4]), 
                        'transcript': str(data[5]), 'assembly_md5': str(data[6]),
                        'proteome_md5': str(data[7]), 'gff3_md5': str(data[8]),
                        'transcript_md5': str(data[9]), 'genome_id': data[10].split(',')
                        }

    return ass_prots

def wait_for_ncbi(count, api = False):
    if count >= 2:
        if not api:
            time.sleep(1)
            count = 0
        elif count >= 7:
            time.sleep(1)
            count = 0
    return count

def esearch_ncbi(accession, column, database = 'assembly'):
    search_term, esc_count = accession + '[' + column + ']', 0
    while esc_count < 10:
        try:
            handle = Entrez.esearch(db=database, term=search_term)
            genome_ids = Entrez.read(handle)['IdList']
            break
        except (RuntimeError, urllib.error.HTTPError) as e:
            time.sleep(1)
            esc_count += 1
    else:
        print('\tERROR:', accession, 'failed to search NCBI')
        return None
    return genome_ids

def esummary_ncbi(ID, database):

    esc_count = 0
    while esc_count < 10:
        esc_count += 1
        try:
            handle = Entrez.esummary(db=database, id=ID, report="full")
            record = Entrez.read(handle, validate = False)
        except urllib.error.HTTPError:
            time.sleep(0.1)
            continue
        if database == 'assembly':
            try: # is it populated with an FTP?
                ftp_path = str(record['DocumentSummarySet']['DocumentSummary'][0]['FtpPath_GenBank'])
            except IndexError: # wait a sec and retry
                time.sleep(1)
                continue
        break
    else: # too many failed attempts
        if esc_count >= 10:
            raise urllib.error.HTTPError('\tERROR: FTP request failed')

    return record

# collects paths to download proteomes and assemblies
def collect_ftps(
    ncbi_df, ass_prots, api_key=0, column = 'assembly_acc',
    ncbi_column='Assembly Accession', database="assembly", output_path = '',
    verbose=True, remove = False, spacer = '\t\t'
    ):

    count, failed = 0, []

# for each row in the assembly, grab the accession number, form the search term for Entrez, use Entrez,
    if ncbi_column in {'assembly', 'genome', 'uid'}:
        out_df = ncbi_df[ncbi_df.index.isin(set(ass_prots.keys()))]
        ncbi_df = ncbi_df[~ncbi_df.index.isin(set(ass_prots.keys()))]
    for accession, row in ncbi_df.iterrows():
        if accession in ass_prots: # add all rows that have indices associated with this
            # query type
            out_df = out_df.append(row)
            icount = 1
            test = str(accession) + '_' + str(icount)
            while test in ass_prots:
                count += 1
    #                row['assembly_acc'] = ass_prots[test]['genome_id']
                if 'ome' in row.keys():
                    row['ome'] = None # haven't assigned a mycotools ID yet
                out_df = out_df.append(row)
                test = str(accession) + '_' + str(icount)
            continue

        elif pd.isnull(row[column]) or not row[column]: # ignore blank entries
            ass_prots[str(accession)] = {
                'accession': accession, 'assembly': '',
                'proteome': '', 'gff3': '', 'transcript': '',
                'assembly_md5': '', 'proteome_md5': '',
                'gff3_md5': '', 'transcript_md5': '', 'genome_id': ''
                }
            failed.append([accession, datetime.strftime(row['version'], '%Y%m%d')])
            continue

        if ncbi_column not in {'uid'}: # we already have the uid, no worries
            genome_id = esearch_ncbi(accession, ncbi_column, database = 'assembly')
        else:
            genome_id = [accession]

        if not genome_id: # No IDs retrieved
            if 'ome' in row.keys():
                accession = row['ome']
            eprint(spacer + '\t' + accession + ' failed to find genome ID', flush = True)
            try:
                failed.append([accession, datetime.strftime(row['version'], '%Y%m%d')])
            except TypeError: # if the row can't be formatted as a date
                failed.append([accession, row['version']])
            continue

        if ncbi_column in {'assembly', 'genome', 'uid'}: # be confident it is the most
        # recent assembly UID
            genome_id = [str(max([int(i) for i in genome_id]))]

        icount = 0
        for ID in genome_id:
            if count:
                new_acc = str(accession) + '$' + str(icount)
            else:
                new_acc = accession
# obtain the path from a summary of the ftp directory and create the standard paths for proteomes and assemblies
            ass_prots[str(new_acc)] = {
                'accession': accession, 'assembly': '', 'proteome': '', 
                'gff3': '', 'transcript': '', 'assembly_md5': '',
                'proteome_md5': '', 'gff3_md5': '', 
                'transcript_md5': '', 'genome_id': ID
                }

            record = esummary_ncbi(ID, database)
            assemblyID = record['DocumentSummarySet']['DocumentSummary'][0]['AssemblyAccession']
            ftp_path = str(record['DocumentSummarySet']['DocumentSummary'][0]['FtpPath_GenBank'])

            if not ftp_path:
                eprint(spacer + '\t' + accession + ' failed to return any FTP path', flush = True)
                failed.append([accession, datetime.strftime(row['version'], '%Y%m%d')])
                continue

            esc_count = 0
            ass_md5, gff_md5, trans_md5, prot_md5, md5s = '', '', '', '', {}
            basename = os.path.basename(ftp_path)
            while True:
                try:
                    esc_count += 1
                    request = requests.get( 
                        (ftp_path + '/md5checksums.txt').replace('ftp://','https://'), timeout = 120
                        )
                    break
                except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
                    if esc_count == 20:
                        eprint(spacer + '\tERROR: Failed to fulfill ftp request!', flush = True)
                        sys.exit(70)
                    count = 0
                    time.sleep(1)
            count += 1
            if request.status_code != 200:
                eprint(spacer + '\t' + ome + '\tno ' + md5, flush = True)
                dwnld = 69
            else:
                dwnld = subprocess.call(
                    ['curl', ftp_path + '/md5checksums.txt', '-o', output_path + '.tmpmd5'], 
                    stdout = subprocess.PIPE, stderr = subprocess.PIPE
                    )
                count += 1
                with open( output_path + '.tmpmd5', 'r' ) as raw:
                    for line in raw:
                        data = line.rstrip().split('  ')
                        if data:
                            md5s[ftp_path + '/' + os.path.basename(data[1])] = data[0]
            tranname = os.path.basename(ftp_path.replace('/GCA','/GCF'))
            assembly = ftp_path + '/' + basename + '_genomic.fna.gz'
            if assembly in md5s:
                ass_md5 = md5s[assembly]
            else:
                assembly = ''
            proteome = ftp_path + '/' + basename + '_protein.faa.gz'
            if proteome in md5s:
                prot_md5 = md5s[proteome]
            else:
                proteome = ''
            gff3 = ftp_path + '/' + basename + '_genomic.gff3.gz'
            test_gff3 = re.sub( r'\.gff3\.gz$', '.gff.gz', gff3 )
            if gff3 in md5s:
                gff_md5 = md5s[gff3]
            elif test_gff3 in md5s:
                gff3 = test_gff3
                gff_md5 = md5s[gff3]
            else:
                gff3 = ''
    
            transcript = ftp_path.replace('/GCA', '/GCF') + '/' + tranname + '_rna.fna.gz'
            if transcript in md5s:
                trans_md5 = md5s[transcript]
    
            if (not assembly or not gff3) and remove:
                failed.append([accession, datetime.strftime(row['version'], '%Y%m%d')])
    
            log_editor( 
                output_path + 'ncbiDwnld.log', str(new_acc), 
                str(accession) + '\t' + accession + '\t' +  assembly + '\t' + \
                proteome + '\t' + gff3 + '\t' + transcript + '\t' + \
                ass_md5 + '\t' + prot_md5 + '\t' + gff_md5 + '\t' + trans_md5 + \
                '\t' + ID
                )
            ass_prots[str(new_acc)] = {
                'accession': accession, 'assembly': assembly, 'assembly_md5': ass_md5,
                'proteome': proteome, 'proteome_md5': prot_md5,
                'gff3': gff3, 'gff3_md5': gff_md5,
                'transcript': transcript, 'transcript_md5': trans_md5,
                'genome_id': ID
                }
            row['dwnld_id'] = ID
            if icount:
                if 'ome' in row:
                    row['ome'] = None
            out_df = out_df.append(row)
            icount += 1
    
    # if no API key is used, we can only generate 3 queries per second, otherwise we can use 10
            count += 1
            count = wait_for_ncbi(count, api_key)
    
    return ass_prots, failed, out_df

# download the file depending on the type inputted
def download_files(acc_prots, acc, file_types, output_dir, count,
                   remove = False, spacer = '\t\t'):

    esc_count, dwnlds = 0, {}
    for file_type in file_types:
        ftp_link = acc_prots[file_type]
        dwnlds[file_type] = -1
        if file_type == 'assembly':
            file_path = output_dir + 'assembly/' + \
                os.path.basename(acc_prots[file_type])
        elif file_type == 'gff3':
            file_path = output_dir + 'gff3/' + \
                os.path.basename(acc_prots[file_type])
        elif file_type == 'proteome':
            file_path = output_dir + 'proteome/' + \
                os.path.basename(acc_prots[file_type])
        elif file_type == 'transcript':
            file_path = output_dir + 'transcript/' + \
                os.path.basename(acc_prots[file_type])

        if os.path.isfile( file_path ):
            count += 1
            md5_cmd = subprocess.run( [
                'md5sum', file_path],
                stdout = subprocess.PIPE )
            md5_res = md5_cmd.stdout.decode( 'utf-8' )
            md5_find = re.search(r'\w+', md5_res)
            md5 = md5_find[0]
            if md5 == acc_prots[file_type + '_md5']:
                eprint(spacer + '\t' + file_type + ': ' + os.path.basename(file_path), flush = True)
                dwnlds[file_type] = 0
                continue
        elif os.path.isfile( file_path[:-3] ):
            eprint(spacer + '\t' + file_type + ': ' + os.path.basename(file_path), flush = True)
            dwnlds[file_type] = 0
            continue

        if ftp_link == '':
            dwnlds[file_type] = 15
            continue
        while True:
            try:
                esc_count += 1
                request = requests.get( ftp_link.replace('ftp://','https://' ))
                break
            except requests.exceptions.ConnectionError:
                if esc_count == 20:
                    eprint(spacer + '\tERROR: Failed to fulfill ftp request!', flush = True)
                    sys.exit(71)
                time.sleep(1)
                count = 0
        count += 1
        if request.status_code != 200:
            eprint(spacer + '\t\t' + file_type + ': ERROR, no file exists', flush = True)
            dwnlds[file_type] = 69
            acc_prots[file_type] = ''
            log_editor( output_dir + 'ncbiDwnld.log', str(acc), str(acc) + \
                '\t' + str(acc_prots['accession']) + '\t' + str(acc_prots['assembly']) + \
                '\t' + str(acc_prots['proteome']) + '\t' + str(acc_prots['gff3']) + \
                '\t' + str(acc_prots['transcript']) + '\t' + \
                str(acc_prots['assembly_md5']) + '\t' + \
                str(acc_prots['proteome_md5']) + '\t' + \
                str(acc_prots['gff3_md5']) + '\t' + \
                str(acc_prots['transcript_md5']) + '\t' + \
                str(','.join([str(x) for x in acc_prots['genacc_id']])))
            if remove and file_type in {'assembly', 'gff3'}:
                break
            continue

        count += 1
        dwnlds[file_type] = subprocess.call(
            ['curl', ftp_link, '-o', file_path], 
            stdout = subprocess.PIPE, stderr = subprocess.PIPE
            )

        if dwnlds[file_type] != 0:
            eprint(spacer + '\t' + file_type + ': ERROR, download failed', flush = True)
            if remove and file_type in {'assembly', 'gff3'}:
                break
        else:
            if os.stat(file_path).st_size < 150:
                eprint(spacer + '\t' + file_type + ': ERROR, file too small', flush = True)
#                       print('\t' + acc + '\t' + file_type + ' empty', flush = True)
                dwnlds[file_type] = 420
                if remove and file_type in {'assembly', 'gff3'}:
                    break
        eprint(spacer + '\t' + file_type + ': ' + os.path.basename(file_path), flush = True)

    return dwnlds, count


def dwnld_mngr(
    data, acc, file_types, output_path, count, remove, api, spacer
    ):
    fail = []
    exits, count = download_files( 
        data, acc, file_types, output_path, 
        count, remove = remove, spacer = spacer
        )
    count = wait_for_ncbi(count, api)
    t_acc = acc
    try:
        if exits['assembly'] != 0:
            if '$' in acc:
                t_acc = acc[:acc.find('$')]
            fail = [t_acc, ncbi_df['version'][t_acc]]
            return fail, count
    except KeyError:
        pass
    try:
        if exits['gff3'] != 0:
            fail = [t_acc, ncbi_df['version'][t_acc]]
            return fail, count
    except KeyError:
        pass
    return fail, count

def dwnld_mngr_no_MD5(
    data, acc, file_types, output_path, count, remove, api, spacer
    ):
    run, fail = False, []
    for file_type in file_types:
        file_path = output_path + file_type \
            + '/' + os.path.basename(data[file_type])
        if not os.path.isfile(file_path):
            run = True
            break
    if run:
        exits, count = download_files( 
            data, acc, file_types, output_path, 
            count, remove = remove,
            spacer = spacer
            )
        count = wait_for_ncbi(count, api)
    else:
        exits = {x: 0 for x in file_types}
        # we aren't checking md5s, so assume the exit is 0
        return fail, count
    try:
        if exits['assembly'] != 0:
            if '$' in acc:
                t_acc = acc[:acc.find('$')]
            else:
                t_acc = acc
            fail = [t_acc, ncbi_df['version'][t_acc]]
            return fail, count
    except KeyError:
        pass
    try:
        if exits['gff3'] != 0:
             fail = [t_acc, ncbi_df['version'][t_acc]]
             return fail, count
    except KeyError:
        pass
    return fail, count


def main( 
    api = None, 
    assembly = True, proteome = False, gff3 = True, transcript = False,
    ncbi_df = False, remove = False, output_path = os.getcwd(), verbose = False,
    column = 'assembly_acc', ncbi_column = 'Assembly', check_MD5 = True,
    spacer = '\t\t'
    ):

    # initialize run directory and information
    output_path = format_path(output_path)
    file_types = prepare_folders( 
        output_path, gff3, proteome, assembly, transcript
        )
    ass_prots = compile_log(output_path, remove )

    # check if ncbi_df is a dataframe, and import if not
    if not isinstance(ncbi_df, pd.DataFrame) and os.path.isfile(ncbi_df):
        ncbi_df = db2df(ncbi_df)
    if len(ncbi_df.index) == 0:
        ncbi_df = pd.DataFrame(
            {i: [v] for i, v in enumerate(list(ncbi_df.keys()))}
            )

    # make the modify date from a standard NCBI table the version if it does
    # not otherwise exist, else there isn't a version to reference
    if 'Modify Date' in ncbi_df.keys() and not 'version' in ncbi_df.keys():
        ncbi_df['version'] = pd.to_datetime(ncbi_df['Modify Date'])
    elif 'version' not in ncbi_df.keys():
        ncbi_df['version'] = ''

    ncbi_df = ncbi_df.set_index(pd.Index(list(ncbi_df[column])))
    # preserve the original column, but index ncbi_df on it as well
    vprint('\n' + spacer + 'Assembling NCBI ftp directories', v = verbose, flush = True)
    ass_prots, failed, ncbi_df = collect_ftps( 
            ncbi_df, ass_prots, remove = remove,
            ncbi_column = ncbi_column, column = column, api_key=api,
            output_path = output_path, verbose = verbose,
            spacer = spacer
            )

    if remove:
        ass_prots = { 
            o: ass_prots[o] for o in ass_prots \
            if all(ass_prots[o][p] for p in ['assembly', 'gff3'])
            }
    new_df = pd.DataFrame()

    vprint('\n' + spacer + 'Downloading NCBI files', v = verbose, flush = True)
    count = 0
    if check_MD5:
        for acc, data in ass_prots.items():
            eprint(spacer + '\t' + str(acc), flush = True)
            if data:
                fail, count = dwnld_mngr(
                    data, acc, file_types, output_path,
                    count, remove, api, spacer
                    )
                if fail:
                    failed.append(fail)
                else:
                    ncbi_df.at[acc, 'assemblyPath'] = output_path + 'assembly/' + \
                        os.path.basename(ass_prots[acc]['assembly'])
                    ncbi_df.at[acc, 'faa'] = output_path + 'proteome/' + \
                        os.path.basename(ass_prots[acc]['proteome'])
                    ncbi_df.at[acc, 'gffPath'] = output_path + 'gff3/' + \
                        os.path.basename(ass_prots[acc]['gff3'])
                    new_df = new_df.append( ncbi_df.loc[acc] )
            else:
                check = ncbi_df[ncbi_df[column] == acc[:acc.find('$')]]
                # check for entries in the inputted table that match the accession
                # provided without version modification
                db_vers = datetime.strftime(row['version'], '%Y%m%d')
                failed.append([acc, db_vers])
    else: # there is no checking md5, this is for efficient, so make it
    # efficient by avoiding conditional expressions
        for acc, data in ass_prots.items():
            eprint(spacer + '\t' + str(acc), flush = True)
            fail, count = dwnld_mngr_no_MD5(
                data, acc, file_types, output_path, count, remove, api, spacer
                )
            if fail:
                failed.append(fail)
            else:
                for file_type in file_types:
                    ncbi_df.at[acc, file_type] = output_path + file_type \
                        + '/' + os.path.basename(data[file_type])
                new_df = new_df.append( ncbi_df.loc[acc] )

    if 'assembly' in new_df.keys():
        new_df = new_df.rename(columns = {'assembly': 'assemblyPath'})
    if 'gff3' in new_df.keys():
        new_df = new_df.rename(columns = {'gff3': 'gffPath'})
    new_df = new_df.reset_index()
    return new_df, failed


def getSRA(assembly_acc, fastqdump = 'fastq-dump', pe = True):

    handle = Entrez.esearch(db='SRA', term=assembly_acc)
    ids = Entrez.read(handle)['IdList']
    for id in ids:
        handle = Entrez.esummary(db='SRA', id = id, report='full')
        records = Entrez.read(handle, validate = False)
        for record in records:
            srr = re.search(r'Run acc="(S\w+\d+)"', record['Runs'])[1]
            print('\t\t' + srr, flush = True)
            if pe:
                subprocess.call([
                    fastqdump, '--gzip', '--split-3', srr], 
                    stdout = subprocess.PIPE)
                if os.path.isfile(srr + '_1.fastq.gz'):
                    shutil.move(srr + '_1.fastq.gz', assembly_acc + '_' + srr + '_1.fq.gz')
                    shutil.move(srr + '_2.fastq.gz', assembly_acc + '_' + srr + '_2.fq.gz')
                else:
                    print('\t\t\tERROR: file failed', flush = True)
            else:
                subprocess.call([
                    fastqdump, '--gzip', srr],
                    stdout = subprocess.PIPE)
                if os.path.isfile(srr + '.fastq.gz'):
                    shutil.move(srr + '.fastq.gz', assembly_acc + '_' + srr + '.fq.gz')
                else:
                    print('\t\t\tERROR: file failed', flush = True)


def goSRA(df, output = os.getcwd() + '/', pe = True):

    print()
    sra_dir = output + 'sra/'
    if not os.path.isdir(sra_dir):
        os.mkdir(sra_dir)
    os.chdir(sra_dir)
    fastqdump = findExecs('fastq-dump', exit = set('fastq-dump'))
    count = 0

    if 'sra' in df.keys():
        row_key = 'sra'
    else:
        row_key = 'assembly_acc'

    for i, row in df.iterrows():
        print('\t' + row[row_key], flush = True)
        getSRA(row[row_key], fastqdump[0])
        count +=1
        if count >= 10:
            time.sleep(1)
            count = 0


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', required = True, \
    help = 'Space delimited accession; tab delimited file with -c')
    parser.add_argument('-a', '--assembly', action = 'store_true')
    parser.add_argument('-p', '--proteome', action = 'store_true')
    parser.add_argument('-g', '--gff3', action = 'store_true')
    parser.add_argument('-t', '--transcript', action = 'store_true')
    parser.add_argument('-s', '--sra', action = 'store_true', \
        help = 'Download SRAs only' )
    parser.add_argument('-pe', '--paired', action = 'store_true', \
        help = 'Download paired-end SRAs. (REQUIRES -s)')
    parser.add_argument('-c', '--column', \
        help = 'Accession column num/name; DEFAULT ["assembly_acc" | 0]')
    parser.add_argument('-n', '--ncbi_column', \
        help = 'NCBI database associated with column. ' \
            + '{"assembly", "biosample", "bioproject", "genome" ...}; ' \
            + 'DEFAULT: attempt to decipher')
    parser.add_argument( '-o', '--output', help = 'Output directory' )
    args = parser.parse_args()

    ncbi_email, ncbi_api, jgi_email, jgi_pwd = loginCheck(jgi = False) 

    Entrez.email = ncbi_email
    if ncbi_api:
        Entrez.api_key = ncbi_api

    if not args.output:
        output = os.getcwd() + '/'
    else:
        output = format_path(args.output)

    args_dict = {
        'NCBI Table': args.input,
        'email': ncbi_email,
        'Assemblies': args.assembly,
        'Proteomes': args.proteome,
        ".gff3's": args.gff3,
        'Transcripts': args.transcript,
        'SRA': args.sra
    }

    start_time = intro('Download NCBI files',args_dict)
    if not args.assembly and not args.proteome and not args.gff3 and not args.sra and not args.transcript:
        eprint('\nERROR: You must choose at least one download option\nExit code 37', flush = True)
        sys.exit( 37 )

    if args.sra:
        if os.path.isfile(format_path(args.input)):
            goSRA(pd.read_csv(format_path(args.input), sep = '\t'), output, pe = args.paired)
        else:
            goSRA(pd.DataFrame({'sra': [args.input.rstrip()]}), output, pe = args.paired)
    else:
        if os.path.isfile(format_path(args.input)):
            ncbi_df = db2df(args.input) # db2df is deprecated, so this needs to change. move out of pandas in general
            if not args.column:
                if 'assembly_acc' in ncbi_df.keys():
                    column = 'assembly_acc'
                    ncbi_column = 'Assembly Accession'
                elif 'Assembly Accession' in ncbi_df.keys():
                    column = 'assembly_acc'
                    ncbi_column = 'Assembly Accession'
                else:
                    column = 0
            try:
                column = ncbi_df.columns[int(column)]
                ncbi_column = column
            except ValueError: # not an integer
                pass
            if not args.ncbi_column:
                if column.lower() in {'assembly'}:
                    ncbi_column = 'assembly'
                elif column.lower() in \
                    {'genome', 'assembly accession', 'assembly_acc'}:
                    ncbi_column = 'genome'
                elif column.lower() in {'biosample', 'biosample accession'}:
                    ncbi_column = 'biosample'
                else:
                    ncbi_column = column.lower()
            else:
                ncbi_column = args.ncbi_column.lower()
        else:
            ncbi_df = pd.DataFrame({'assembly_acc': args.input.replace('"','').replace("'",'').split()})
            column = 'assembly_acc'
            ncbi_column = 'assembly'
    
        new_df, failed = main( 
            assembly = args.assembly, column = column, 
            ncbi_column = ncbi_column, proteome = args.proteome, 
            gff3 = args.gff3, transcript = args.transcript, ncbi_df = ncbi_df,
            output_path = output, verbose = True, spacer = ''
            )
        new_df.to_csv( args.input + '_dwnld', sep = '\t' )

    outro(start_time)



import setuptools

with open( "README.md", "r" ) as fh:
    long_description = fh.read()

setuptools.setup(
    name = "mycotools",
    version = "0.26.25",
    author = "xonq",
    author_email = "konkelzach@protonmail.com",
    description = "A compilation of bioinformatic and computation biology inspired " + \
        "python tools",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://gitlab.com/xonq/mycotools/mycotools",
    package_dir={"": "mycotools"},
    packages = setuptools.find_packages( where="mycotools" ),
    scripts = ['mycotools/mycotools/curAnnotation.py', 'mycotools/mycotools/acc2fa.py', 'mycotools/mycotools/acc2gff.py', 'mycotools/mycotools/fa2clus.py', 'mycotools/mycotools/annotationStats.py', 'mycotools/mycotools/assemblyStats.py', 'mycotools/mycotools/extractDB.py', 'mycotools/mycotools/fa2mass.py', 'mycotools/mycotools/gff2svg.py', 'mycotools/mycotools/crap.py', 'mycotools/mycotools/db2hmm.py', 'mycotools/mycotools/acc2locus.py', 'mycotools/mycotools/fa2tree.py', 'mycotools/mycotools/ncbiDwnld.py', 'mycotools/mycotools/ncbiAcc2fa.py', 'mycotools/mycotools/jgiDwnld.py', 'mycotools/mycotools/ome2name.py', 'mycotools/mycotools/coords2fa.py', 'mycotools/mycotools/gff2seq.py', 'mycotools/mycotools/predb2db.py', 'mycotools/mycotools/acc2gbk.py', 'mycotools/mycotools/db2search.py', 'mycotools/mycotools/fa2hmmer2fa.py', 'mycotools/mycotools/db2files.py', 'mycotools/mycotools/mtdb', 'mycotools/mycotools/utils/curGFF3.py'],
    install_requires = ['biopython', 'pandas', 'requests', 'scipy', 'openpyxl'],
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires = '>=3.0,<4'
)



#! /usr/bin/env python3

import os
import re
import sys
import argparse
import multiprocessing as mp
from collections import defaultdict
from Bio.Seq import Seq
from mycotools.lib.kontools import eprint, format_path, stdin2str
from mycotools.lib.dbtools import mtdb, masterDB
from mycotools.lib.biotools import fa2dict, gff2list, gff3Comps
from mycotools.acc2gff import dbMain as acc2gff

def col_CDS(gff_list, types = {'gene', 'CDS', 'exon', 'mRNA',
                               'tRNA', 'rRNA', 'RNA', 'pseudogene'}):
    '''Collect all CDS entries from a `gff` and store them into cds_dict. If the gene found in the CDS
    is in the set of `ids` then add that type to the CDS dict for that protein
    header.'''

    cds_dict = defaultdict(dict)
    for entry in gff_list:
        if entry['type'] in types:
            contig = entry['seqid']
            try:
                alias = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
            except TypeError:
                eprint( '\n\tERROR: could not extract Alias ID from ' + gff , flush = True)
                continue
            aliases = alias.split('|') # to address alternate splicing in gene
            # aliases
            for alias in aliases:
                ome = alias[:alias.find('_')]
                if contig not in cds_dict[ome]:
                    cds_dict[ome][contig] = defaultdict(list)
                cds_dict[ome][contig][alias].append(entry)
            
    return cds_dict

def contig2gbk(ome, row, contig, contig_dict, 
               contig_seq, faa, product_search = r'product=([^;]+)'):
    """Generate a genbank string for a contig_dict, which contains a dictionary
    of protein keys and list of each gff entry associated with the protein. 
    ome: ome_code
    row: mtdb row
    contig: contig name
    contig_dict: {prot: [gff_entries]}
    contig_seq: contig sequence string
    faa: proteome
    product_search: product regular expression"""

    seq_coords = [] # a list of all the coordinates from the contig
    for prot, prot_list in contig_dict.items():
        for entry in prot_list:
            t_start, t_end = int(entry['start']), int(entry['end'])
            seq_coords.append(sorted([t_start, t_end]))

    # sort the coordinates
    seq_coords.sort(key = lambda x: x[0])
    # the starting index
    init_start_test = int(seq_coords[0][0]) - 1
    init_end_test = int(seq_coords[-1][1])
    end_test = init_end_test - init_start_test # relative end
    start_test = 1 # relative start

    # specify the contig edge - may cause parsers (BiG-SCAPE/antiSMASH) to fail
    edge = '/contig_edge=' + str(init_start_test) + '-;' \
         + str(len(contig_seq) - init_end_test) + '+'
    name = ome + '_' + contig
    relative_end = seq_coords[-1][1] - seq_coords[0][0]
    gbk = "LOCUS       " + name + '\n' + \
        "DEFINITION  " + name + '\n' + \
        "ACCESSION   " + contig + '\n' + \
        "VERSION     " + name + '\n' + \
        "KEYWORDS    .\nSOURCE    " + row['source'] + "\n  ORGANISM  " + \
        str(row['genus']) + '_' + str(row['species']) + '_' + \
        str(row['strain']) + '_' + ome + '\n' + \
        "COMMENT     coordinates are relative to adjustment: " + str(init_start_test) + \
        '\n            .\nFEATURES             Location/Qualifiers\n' + \
        '     region          ' + str(start_test) + '..' + str(end_test) + '\n' + \
        '                     ' + str(edge) + '\n'
#        '     region          1..' + str(relative_end) + '\n' + \
        # this should be
        # related to the start of the contig

    used_aliases = set() # to address alternate splicing and 1 entry per gene
    for prot, prot_list in contig_dict.items():
        final_coords = ''
        entries = defaultdict(list)
        for entry in prot_list:
            if entry['type'].lower() == 'cds':
                entries['CDS'].append(entry)
            elif entry['type'].lower() == 'exon':
                entries['exon'].append(entry)
            elif entry['type'].lower() in {'mrna', 'rrna', 'trna', 'rna'}:
                entries['RNA'].append(entry)
            elif entry['type'].lower() in {'gene', 'pseudogene'}:
                entries['gene'].append(entry)


        for entry in entries['gene']:
            alias = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
            if '|' in alias: # alternately spliced gene
                if alias in used_aliases:
                    continue
                else:
                    used_aliases.add(alias)
            id_ = re.search(gff3Comps()['id'], entry['attributes'])[1]

            try:
                product = re.search(product_search, entry['attributes'])[1]
            except TypeError: # no product
                product = ''
            start, end = sorted([int(entry['start']), int(entry['end'])])
            gbk += '     ' + entry['type'] + \
                '                '[:-len(entry['type'])]
            if entry['strand'] == '+':
                gene_coords = str(start-init_start_test) + '..' \
                    + str(end-init_start_test) + '\n                     '
            else:
                gene_coords = 'complement(' + str(start-init_start_test) \
                    + '..' + str(end-init_start_test) + ')\n                     '
            gbk += gene_coords
            gbk += '/locus_tag="' + alias + '"\n                     ' 

            if product:
                gbk += '/product="' + product + '"\n                     '
            if entry['phase'] != '.':
                gbk += '/phase="' + entry['phase'] + '"\n                     '
            gbk += '/source="' + entry['source'] + '"\n'

        for entry in entries['RNA']:
            gbk += '     ' + entry['type'] + \
                '                '[:-len(entry['type'])]
            if final_coords:
                gbk += final_coords
            else:
                gbk += gene_coords
            gbk += '/locus_tag="' + alias + '"\n                     ' 

            if product:
                gbk += '/product="' + product + '"\n                     '
            if entry['phase'] != '.':
                gbk += '/phase="' + entry['phase'] + '"\n                     '
            gbk += '/source="' + entry['source'] + '"\n'
            gbk += '                     /transcript_id="' + id_ + '"\n'

        if entries['CDS']:
            entry = entries['CDS'][0]
            strand = entry['strand']
            cds_coords = ''
            if strand == '+':
                for cds in entries['CDS']:
                    coords = [cds['start'] - init_start_test, 
                              cds['end'] - init_start_test]
                    max_c, min_c = max(coords), min(coords)
                    cds_coords += str(min_c) + '..' + str(max_c) + ','
            else:
                for cds in entries['CDS']:
                    coords = [cds['start'] - init_start_test, 
                              cds['end'] - init_start_test]
                    max_c, min_c = max(coords), min(coords)
                    cds_coords += 'complement(' + str(min_c) + '..' \
                               + str(max_c) + '),'
            cds_coords = cds_coords[:-1]
            cds_coords_list = cds_coords.split(',')
            final_coords = '' 
            total_len = 0
            for coord in cds_coords_list[:-1]:
                if total_len + len(coord) < 57 or total_len == 0:
                    final_coords += coord + ','
                    total_len += len(coord) + 1
                else:
                    total_len = len(coord) + 1
                    final_coords += '\n                     ' \
                                 + coord + ','
            if total_len == 0 or total_len + len(cds_coords_list[-1]) < 58:
                final_coords += cds_coords_list[-1]
            else:
                final_coords += '\n                     ' \
                             + cds_coords_list[-1]
            if len(cds_coords_list) > 1:
                final_coords = 'join(' + final_coords + ')'
            alias = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
            id_ = re.search(gff3Comps()['id'], entry['attributes'])[1]

            try:
                product = re.search(product_search, entry['attributes'])[1]
            except TypeError: # no product
                product = ''
            gbk += '     ' + entry['type'] + \
                '                '[:-len(entry['type'])]
            final_coords += '\n                     '
            gbk += final_coords
            gbk += '/locus_tag="' + alias + '"\n                     ' 

            if product:
                gbk += '/product="' + product + '"\n                     '
            if entry['phase'] != '.':
                gbk += '/phase="' + entry['phase'] + '"\n                     '
            gbk += '/source="' + entry['source'] + '"\n'
            gbk += '                     '
            gbk += '/protein_id="' + alias + '"\n'
            gbk += '                     ' + \
                '/transl_table=1\n                     /translation="'
            inputSeq = faa[alias]['sequence']
            lines = [inputSeq[:45]]
            if len(inputSeq) > 45:
                restSeq = inputSeq[45:]
                n = 59
                toAdd = [restSeq[i:i+n] for i in range(0, len(restSeq), n)]
                lines.extend(toAdd)
            gbk += lines[0] + '\n'
            if len(lines) > 1:
                for line in lines[1:len(lines)-1]:
                    gbk += '                     ' + line + '\n'
                gbk += '                     ' + lines[-1] + '"\n'
            else:
                gbk += '"\n'

    total_coords = []
    for coord in seq_coords:
        total_coords.extend(coord)
    min_c, max_c = min(total_coords), max(total_coords)
#    for coordinates in seq_coords:
    assSeq = contig_seq[min_c:max_c]
    seqLines = [assSeq[i:i+60] for i in range(0, len(assSeq), 60)]
    count = -59

    gbk += 'ORIGIN\n'
    for line in seqLines:
        count += 60
        gbk += '{:>9}'.format(str(count)) + ' '
        seps = [line[i:i+10] for i in range(0, len(line), 10)]
        for sep in seps:
            gbk += sep.lower() + ' '
        gbk = gbk.rstrip() + '\n'
    
    gbk += '//'
    return gbk

def genGBK(ome_dict, row, ome, product_search):

    assembly = fa2dict(row['fna'])
    faa = fa2dict(row['faa'])
    gbk = {}
    for contig, contig_dict in ome_dict.items():
        gbk[contig] = contig2gbk(ome, row, contig, contig_dict, 
                   assembly[contig]['sequence'], faa, 
                   product_search = product_search)

    return gbk
        

def main(gff_list, db, product_search = r'product=([^;]+)'):
    cds_dict = col_CDS(gff_list, 
            types = {'gene', 'CDS', 'exon', 'mRNA', 
                     'tRNA', 'rRNA', 'RNA', 'pseudogene'})
    ome_gbks = {}
    for ome, ome_dict in cds_dict.items():
        ome_gbks[ome] = genGBK(ome_dict, db[ome], ome, product_search)

    return ome_gbks

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description = "Inputs MTDB gff or accessions, outputs GenBank file" )
    parser.add_argument('-a', '--accession', help = '"-" for stdin')
    parser.add_argument('-i', '--input', help = 'File with accessions')
    parser.add_argument('-g', '--gff', help = 'GFF3 input')
    parser.add_argument('-p', '--product', 
        help = 'Product regular expression: DEFAULT: "Product=([^;]+)"')
    parser.add_argument('-o', '--ome', action = 'store_true', 
        help = 'Output files by ome code')
    parser.add_argument('-s', '--seqid', action = 'store_true',
        help = 'Output files by sequence ID (contig/scaffold/chromosome)'
        )
    parser.add_argument('-d', '--mtdb',
        help = "DEFAULT: master", default = masterDB())
    parser.add_argument('-c', '--cpu', type = int, default = 1)
    args = parser.parse_args()

    if args.input:
        input_file = format_path(args.input)
        with open(input_file, 'r') as raw:
            accs_prep = [x.rstrip().split() for x in raw]
        accs = []
        for acc in accs_prep:
            accs.extend(acc)
    elif args.accession:
        if args.accession == '-':
            data = stdin2str()
            accs = data.split()
        else: 
            accs = [args.accession]
    elif not args.gff:
        raise FileNotFoundError('need input file or accession')

    if not args.product:
        regex = r'Product=([^;])+'
    else:
        regex = r''
        if args.regex.startswith(("'",'"')):
            args.regex = args.regex[1:]
        if args.regex.endswith(('"',"'")):
            args.regex = args.regex[:-1]
        for char in args.regex:
            regex += char

    # import database and set index
    db = mtdb(format_path(args.mtdb))
    db = db.set_index() 

    gbk_dict = {}
    if args.gff:
        gbk_dict = main(gff2list(format_path(args.gff)), db, regex)
    else:
        gffs = acc2gff(db, accs, args.cpu)
        for ome, gff in gffs.items():
            gbk_dict[ome] = main(gff, db, regex)[ome]
    
    if args.ome:
        for ome, gbks in gbk_dict.items():
            with open(ome + '.acc2gbk.gbk', 'w') as out:
                out.write('\n'.join([gbk for gbk in gbks]))
    elif args.seqid:
        for ome, gbks in gbk_dict.items():
            for contig, gbk in gbks.items():
                with open(ome + '_' + contig + '.gbk', 'w') as out:
                    out.write(gbk)
    else:
        for ome, gbks in gbk_dict.items():
            for contig, gbk in gbks.items():
                print(gbk, flush = True)

    sys.exit(0)



#! /usr/bin/env python3

# NEED to come up with more clever way of detecting mtdb

'''
Takes a database as argument 1 and an output for the .tsv as argument 2.
Calculates basic genome statistics.
'''

import os
import sys
import copy
import multiprocessing as mp
from mycotools.lib.dbtools import mtdb
from mycotools.lib.biotools import fa2dict
from mycotools.lib.kontools import format_path, eprint


def calcMask( contig_list ):

    seq = ''.join([x['sequence'] for x in contig_list])
    mask = seq.count('a')
    mask += seq.count('t')
    mask += seq.count('g')
    mask += seq.count('c')

    return mask


def sortContigs( assembly_path ):
    '''Imports fasta, creates a list of dicts for each contig length and its name.
       Sorts the list in descending order by length'''

    contigList = []
    assembly = fa2dict( assembly_path )
    for contig in assembly:
        contigList.append( {
            'len': len(assembly[contig]['sequence']), 
            'name': contig, 
            'sequence': str(assembly[contig]['sequence'])
        } )

    sortedList = sorted(contigList, key = lambda i: i['len'], reverse = True)

    return sortedList


def n50l50( sortedContigs ):
    '''Calculates contigs greater than 1000 bp, the total length excluding 1000 bp contigs.
       Then caculates l50, n50, largest contig, total contigs, shortest contig, and l50%.
       Returns a dictionary of the results.'''

    pass_fa = []
    total, total1000, bp1000, gc, gc1000, gctot, gctot1000 = 0, 0, 0, 0, 0, 0, 0
    for contig in sortedContigs:
        total += contig['len']
        gc += contig['sequence'].lower().count( "g" ) + contig['sequence'].lower().count( 'c' )
        gctot += contig['sequence'].lower().count( "g" ) + contig['sequence'].lower().count( 'c' ) + \
           contig['sequence'].lower().count( 'a' ) + contig['sequence'].lower().count( 't' ) 
        if contig['len'] >= 1000:
            gc1000 += contig['sequence'].lower().count( "g" ) + contig['sequence'].lower().count( 'c' )
            gctot1000 += contig['sequence'].lower().count( "g" ) + contig['sequence'].lower().count( 'c' ) + \
                contig['sequence'].lower().count( 'a' ) + contig['sequence'].lower().count( 't' ) 
            total1000 += contig['len']
            bp1000 += 1
            pass_fa.append( contig )


    out = {}
    count50, count, check = 0, 0, 0
    for contig in sortedContigs:
        count50 += contig['len']
        count += 1
        if count50 >= total1000/2:
            if contig['len'] >= 1000 and check != 1:
                out['n50-1000bp'] = int(contig['len'])
                out['l50-1000bp'] = int(count)
                out['l50%-1000bp'] = out['l50-1000bp']/int(bp1000)
                check = 1
            if count50 >= total/2:
                out['n50'] = int(contig['len'])
                out['l50'] = int(count)
                break

    try:
        out['l50%'] = out['l50']/len(sortedContigs)
        out['largest_contig'] = sortedContigs[0]['len']
        out['shortest_contig'] = sortedContigs[-1]['len']
        out['contigs'] = len(sortedContigs)
        out['contigs-1000bp'] = bp1000
        out['assembly_len'] = int(total)
        out['assembly_len-1000bp'] = int(total1000)
        out['gc'] = float(gc/gctot)
        out['gc-1000bp'] = float(gc1000/gctot1000)
        if 'n50-1000bp' not in out:
            out['n50-1000bp'] = 'na'
            out['l50-1000bp'] = 'na'
        maskCount = calcMask( sortedContigs )
        maskCount1000 = calcMask( pass_fa )
        out['mask%'] = maskCount / int( total ) * 100
        out['mask%-1000bp'] = maskCount1000 / int( total1000) * 100
        
    except KeyError:
        out = {}


    return out

def mngr(assembly_path, ome):
    sortedContigs = sortContigs(assembly_path)
    calcs = n50l50(sortedContigs)
    return ome, tuple([(x, calcs[x]) for x in calcs])


def main(in_path, log_path = None, cpus = 1):

    stats = {}

    if in_path.endswith('db'):
        head = '#ome\tn50-1000bp\tl50-1000bp\tl50%-1000bp\tn50\tl50\tl50%\tlargest_contig\tshortest_contig\tcontigs' + \
            '\tcontigs-1000bp\tassembly_len\tassembly_len-1000bp\tgc\tgc-1000bp\tmask%\tmask%-1000bp'

        prevOmes = {}
        if log_path:
            if not os.path.isfile( log_path ):
                with open( log_path, 'w' ) as log_open:
                    log_open.write( head )
            else:
                with open(log_path, 'r') as raw:
                    for line in raw:
                        if not line.startswith('#'):
                            omeI = line.index('\t')
                            ome = line[:omeI]
                            prevOmes[ome] = line[omeI+1:].rstrip()

        db = mtdb(in_path).set_index()

        cmds = []
        for ome in db:
            row = db[ome]
            if row['fna'] and ome not in prevOmes:
                cmds.append((row['fna'], ome,))
        with mp.Pool(processes=cpus) as pool:
            results = pool.starmap(mngr, cmds)

        calcs = {}
        for res in results:
            if res[1]:
                calcs[res[0]] = '\t'.join([str(x[1]) for x in res[1]])
            else:
                eprint('\t\tERROR:\t' + ome, flush = True)

        calcs = {
            k: v for k,v in \
            sorted(
                {**calcs, **prevOmes}.items(), 
                key = lambda x: x[0]
                )
            }

        if log_path:
            with open(log_path, 'w') as logWrite:
                logWrite.write(head + '\n')
                for ome in calcs:
                    data = calcs[ome]
                    logWrite.write(
                        ome + '\t' + data + '\n'
                        )
        else:
            print(head, flush = True)
            for ome in calcs:
                data = calcs[ome]
                print(
                    ome + '\t' + data, flush = True
                    )

    else:
        sortedContigs = sortContigs(in_path)
        calculations = n50l50( sortedContigs )
        if calculations:
            stats[ os.path.basename( os.path.abspath(in_path)) ] = n50l50( sortedContigs )
        else:
            eprint('\tERROR:\t' + in_path, flush = True)

        for stat in stats:
            if stats[stat]['shortest_contig'] >= 1000:
                stats[stat] = { 
                    info: stats[stat][info]  for info in stats[stat] if '1000bp' not in info 
                }
            for info in stats[stat]:
                print( '{:<25}'.format( info.upper() + ':' ) + str( stats[stat][info] ) , flush = True)
            

if __name__ == '__main__':
    usage = '\nUSAGE: assembly statistics\nAssembly `fasta` or mycotoolsDB, optional output file if using database\n'
    if {'-h', '--help'}.intersection(set(sys.argv)):
        print(usage, flush = True)
        sys.exit(1)
    if len(sys.argv) < 2:
        print( usage , flush = True)
        sys.exit( 1 )

    in_path = format_path(sys.argv[1])
    if len(sys.argv) > 2:
        log_path = format_path(sys.argv[2])
    else:
        log_path = None

    main(in_path, log_path, os.cpu_count())

    sys.exit(0)

